import queue
import weakref
import logging
import threading
from concurrent import futures
from contextlib import contextmanager
from typing import Optional, Callable, Set, Generator, Any, Tuple

from ... import errors

logger = logging.getLogger(__name__)

ExecuteType = Callable[[callable, tuple, dict, futures.Future], Any]
WorkerExitMessageType = Callable[[], str]
ExecuteContextReturnType = Generator[
    Tuple[ExecuteType, WorkerExitMessageType], None, None
]


class ExecutorShutdown(Exception):
    pass


class RemoteExecutor(futures.Executor):
    """Asynchronous executor which delegates task execution to :math:`WorkerProxy` instances."""

    _FUTURE_CLASS = futures.Future

    def __init__(
        self,
        max_workers: Optional[int] = None,
        max_tasks_per_worker: Optional[int] = None,
        lazy_scheduling: bool = True,
        conservative_scheduling: bool = False,
    ):
        """
        :param max_workers: maximum number of workers that can run at any given time. `None` means unlimited.
        :param max_tasks_per_worker: maximum number of tasks each worker can receive before exiting. `None` means unlimited.
        :param lazy_scheduling: schedule workers only when needed. Can only be disabled when `max_workers` is specified.
        :param conservative_scheduling: schedule the least amount of workers at the expense of tasks staying longer in the queue.
        """
        if max_workers is None:
            if not lazy_scheduling:
                raise ValueError(
                    "Cannot disable lazy scheduling when there is no maximum number of workers"
                )
        else:
            if not isinstance(max_workers, int):
                raise TypeError("max_workers must be an integer")
            elif max_workers <= 0:
                raise ValueError("max_workers must be greater than 0")
        self._max_workers = max_workers

        if max_tasks_per_worker is not None:
            if not isinstance(max_tasks_per_worker, int):
                raise TypeError("max_tasks_per_worker must be an integer")
            elif max_tasks_per_worker <= 0:
                raise ValueError("max_tasks_per_worker must be >= 1")
        self._max_tasks_per_worker = max_tasks_per_worker

        self._lazy_scheduling = lazy_scheduling
        self._task_queue = queue.Queue()
        self._shutdown_lock = threading.Lock()
        self._workers: Set[threading.Thread] = set()
        self._futures = weakref.WeakSet()
        self._shutdown_flag = False
        self._conservative_scheduling = conservative_scheduling

        if not lazy_scheduling:
            for _ in range(self._max_workers):
                self._start_worker()

    def submit(self, task: Callable, *args, **kwargs) -> futures.Future:
        """
        :param task: function to be executed remotely
        :param args: positional arguments of the function to be executed remotely
        :param kwargs: named arguments of the function to be executed remotely
        :returns: future object to retrieve the result
        """
        with self._shutdown_lock:
            if self._shutdown_flag:
                raise RuntimeError("ThreadPool has been shut down")

            future = self._FUTURE_CLASS()
            # self._futures.add(future)
            self._task_queue.put((task, args, kwargs, future))
            logger.debug(
                "%s: submitted (task capacity = %s)",
                type(self).__name__,
                self._task_capacity(),
            )

            if self._lazy_scheduling and self._require_more_workers():
                self._start_worker()
        return future

    def shutdown(self, wait: bool = True, cancel: bool = False):
        """
        :param wait: wait for all workers to exit
        :param cancel: cancel all pending tasks (running once cannot be cancelled)
        """
        with self._shutdown_lock:
            self._shutdown_flag = True
            workers = list(self._workers)

            if cancel:
                for future in list(self._futures):
                    future.cancel()

            # Sentinel to shut down workers
            for _ in range(len(workers)):
                self._task_queue.put((None, None, None, None))

        if wait:
            for worker in workers:
                worker.join()
        self._workers.clear()

    def _task_capacity(self) -> int:
        """The total number of tasks that the running workers can execute."""
        return sum(worker.task_capacity for worker in self._workers)

    def _idle_workers(self) -> int:
        """The total number of idle workers"""
        return sum(worker.idle for worker in self._workers)

    def _max_task_queue_size(self) -> int:
        if self._conservative_scheduling:
            return self._task_capacity()
        else:
            return self._idle_workers()

    def _require_more_workers(self) -> bool:
        if self._shutdown_flag:
            # Do not start workers when shutting down
            return False
        if self._max_workers and len(self._workers) >= self._max_workers:
            # Already the maximum amount of workers
            return False
        if self._task_queue.qsize() <= self._max_task_queue_size():
            # Workers have enough capacity to drain the task queue
            return False
        return True

    def _start_worker(self) -> None:
        if self._shutdown_flag:
            # Do not start workers when shutting down
            return False

        used_indices = {w.idx for w in self._workers}
        idx = 0
        while idx in used_indices:
            idx += 1

        thread_name = f"{type(self).__name__}-{idx}"
        task_capacity = self._max_tasks_per_worker or float("inf")
        worker = threading.Thread(
            target=self._worker_main,
            name=thread_name,
            daemon=True,
        )
        worker.task_capacity = task_capacity
        worker.idle = False
        worker.idx = idx
        worker.start()
        self._workers.add(worker)
        return True

    def _worker_main(self):
        worker = threading.current_thread()
        logger.info(
            "%s: started (task capacity = %s)", worker.name, worker.task_capacity
        )
        exit_reason = "cancelled"
        try:
            with self.execute_context() as (execute, worker_exit_msg):
                while not self._shutdown_flag:
                    try:
                        worker.idle = True
                        task, args, kwargs, future = self._task_queue.get(timeout=1)
                    except queue.Empty:
                        exit_msg = worker_exit_msg()
                        if exit_msg:
                            exit_reason = exit_msg
                            break
                        continue

                    try:
                        worker.idle = False

                        if task is None:  # Sentinel to shut down the worker
                            exit_reason = "requested to stop"
                            break

                        worker.task_capacity -= 1
                        if future.set_running_or_notify_cancel():
                            logger.info(
                                "%s: start executing %s (task capacity = %s)",
                                worker.name,
                                task,
                                worker.task_capacity,
                            )
                            try:
                                result = execute(task, args, kwargs, future)
                            except Exception as exc:
                                future.set_exception(exc)
                                # self = None  # TODO: break a reference cycle with the exception 'exc' but we need self later
                                if isinstance(
                                    exc, (ExecutorShutdown, errors.RemoteExit)
                                ):
                                    exit_reason = str(exc)
                                    logger.info(
                                        "%s: failed executing %s — %s (task capacity = %s)",
                                        worker.name,
                                        task,
                                        exit_reason,
                                        worker.task_capacity,
                                    )
                                    break
                                else:
                                    logger.error(
                                        "%s: failed executing %s (task capacity = %s)",
                                        worker.name,
                                        task,
                                        worker.task_capacity,
                                        exc_info=True,
                                    )
                            else:
                                future.set_result(result)
                                logger.info(
                                    "%s: succeeded executing %s (task capacity = %s)",
                                    worker.name,
                                    task,
                                    worker.task_capacity,
                                )
                        else:
                            logger.warning(
                                "%s: future cancelled (task capacity = %s)",
                                worker.name,
                                worker.task_capacity,
                            )
                    finally:
                        self._task_queue.task_done()

                    if worker.task_capacity <= 0:
                        exit_reason = "task capacity reached"
                        break
        except BaseException:
            logger.critical("Exception in worker", exc_info=True)
            raise

        try:
            logger.debug("%s: %s", worker.name, exit_reason)
            with self._shutdown_lock:
                self._workers.discard(worker)
                if not self._shutdown_flag:
                    if self._lazy_scheduling:
                        while self._require_more_workers():
                            self._start_worker()
                    else:
                        self._start_worker()
            logger.debug("%s: exiting", worker.name)
        except BaseException:
            logger.critical("Exception in worker exit", exc_info=True)
            raise

    @contextmanager
    def execute_context(
        self,
    ) -> ExecuteContextReturnType:
        def execute(
            task: callable, args: tuple, kwargs: dict, future: futures.Future
        ) -> Any:
            return task(*args, **kwargs)

        def worker_exit_msg() -> Optional[str]:
            """Return exit message in case the executor worker must exit."""
            return

        yield (execute, worker_exit_msg)



import abc
import pickle
import logging
import threading
from contextlib import contextmanager
from typing import Generator, Tuple, Any, Optional


logger = logging.getLogger(__name__)


class Connection(abc.ABC):
    """Server-side of a connection. The client-side is `ewoksjob.client.job_io.remote._base.Client`."""

    _HEADER_NBYTES = 64  # allows for 2**64 bytes (~18 EB) data transfer
    _MAX_CHUNK_NBYTES = 65536  # data transfer in chunks of 64 KB

    def __init__(self, raise_on_status_error: Optional[callable] = None) -> None:
        """
        :param raise_on_status_error: function that raises an exception when there is a status error (remote or local exit).
        """
        self._cancel_event = threading.Event()
        self._yield_period = 1
        self._raise_on_status_error = raise_on_status_error

    @property
    @abc.abstractmethod
    def input_name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def output_name(self) -> str:
        pass

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def cancel(self) -> None:
        self._cancel_event.set()

    def cancelled(self) -> bool:
        return self._cancel_event.is_set()

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def _wait_client(self) -> None:
        """Wait for the client to be online. The client is remote, we are the server.

        :raises RemoteExit: raises exception when client is not alive
        """
        pass

    @contextmanager
    def _wait_client_context(self) -> Generator[None, None, None]:
        logger.debug("waiting for remote job to connect to %s ...", self.output_name)
        try:
            yield
        except Exception:
            logger.debug(
                "waiting for remote job to connect to %s failed", self.output_name
            )
            raise
        if self.cancelled():
            logger.debug(
                "waiting for remote job to connect to %s cancelled", self.output_name
            )
        else:
            logger.debug("remote job connected to %s", self.output_name)

    def send_data(self, data: Any) -> None:
        """Send data to the client."""
        bdata = self._serialize_data(data)
        nbytes = len(bdata)
        logger.debug(
            "send data %s (%d bytes) to client of %s ...",
            type(data),
            nbytes,
            self.output_name,
        )
        bheader = self._serialize_header(bdata)
        try:
            self._send_bytes_with_check(bheader + bdata)
        except (BrokenPipeError, ConnectionResetError):
            if data is None:
                logger.debug("client of %s already exited", self.output_name)
                return
            raise
        logger.debug("data send to client of %s", self.output_name)

    def receive_data(self) -> Tuple[Any, Optional[BaseException]]:
        """Receive data from the client.

        :raises RemoteExit: raises exception when client is not alive
        """
        logger.debug("waiting for client data on %s ...", self.input_name)
        bheader = self._receive_nbytes_with_check(self._HEADER_NBYTES)
        nbytes = self._deserialize_header(bheader)
        if nbytes == 0:
            logger.warning(
                "corrupt header %s from client on %s ...", bheader, self.input_name
            )
        bdata = self._receive_nbytes_with_check(nbytes)
        logger.debug("client data received from %s", self.input_name)
        return self._deserialize_data(bdata)

    def _serialize_header(self, bdata: bytes) -> bytes:
        return len(bdata).to_bytes(self._HEADER_NBYTES, "big")

    def _deserialize_header(self, bheader: bytes) -> int:
        return int.from_bytes(bheader, "big")

    def _serialize_data(self, data: Any) -> bytes:
        return pickle.dumps(data)

    def _deserialize_data(self, data: bytes) -> Any:
        return pickle.loads(data)

    def _receive_nbytes_with_check(self, nbytes: int) -> bytes:
        """
        :raises RemoteExit: raises exception when not alive
        :raises ValueError: cancelled or did not receive the requested number of bytes
        """
        bdata = self._receive_nbytes(nbytes)
        if len(bdata) != nbytes:
            err_msg = f"{len(bdata)} bytes received from {self.input_name} instead of {nbytes} bytes"
            if self.cancelled():
                raise ValueError(f"{err_msg} (cancelled)")
            else:
                raise ValueError(err_msg)
        return bdata

    @abc.abstractmethod
    def _receive_nbytes(self, nbytes: int) -> bytes:
        """Receive exactly nbytes from the client.

        :raises RemoteExit: raises exception when client is not alive
        """
        pass

    def _send_bytes_with_check(self, data: bytes) -> None:
        """
        :raises RuntimeError: cancelled
        """
        if self.cancelled():
            raise RuntimeError(f"send to %{self} is cancelled")
        self._send_bytes(data)

    @abc.abstractmethod
    def _send_bytes(self, data: bytes) -> None:
        """Send bytes to the client."""
        pass


class Buffer:
    def __init__(self, nbytes: int, max_chunk_size: int) -> None:
        self._data = bytearray()
        self._nbytes = nbytes
        self._remaining = nbytes
        self._has_data = False
        self._MAX_CHUNK_NBYTES = max_chunk_size

    @property
    def read_size(self) -> int:
        return min(self._remaining, self._MAX_CHUNK_NBYTES)

    def extend(self, bdata: bytes) -> None:
        if bdata:
            self._data.extend(bdata)
            self._remaining -= len(bdata)
            self._has_data = True

    @property
    def complete(self) -> bool:
        return self._remaining <= 0

    @property
    def has_data(self) -> bool:
        return self._has_data

    @property
    def progress(self) -> str:
        return f"{len(self._data)}/{self._nbytes} bytes"

    def to_bytes(self) -> bytes:
        return bytes(self._data)



def add_parameters(parser):
    parser.add_argument(
        "jobid1",
        type=str,
        help="SLURM Job ID",
    )
    parser.add_argument(
        "jobid2",
        type=str,
        help="SLURM Job ID",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="All users when '--jobid' is not set",
    )
    parser.add_argument(
        "-m",
        "--monitor",
        dest="interval",
        type=float,
        default=0,
        help="Monitor are intervals of x seconds",
    )


def apply_parameters(args):
    pass



from typing import Mapping, Optional


def merge_mappings(d1: Optional[Mapping], d2: Optional[Mapping]) -> dict:
    """Same as `{**d1, **d2}` but then recursive"""
    if d1 is None:
        merged = dict()
    else:
        merged = dict(d1)
    if not d2:
        return merged
    for key, value2 in d2.items():
        value1 = merged.get(key)
        if isinstance(value1, Mapping) and isinstance(value2, Mapping):
            value2 = merge_mappings(value1, value2)
        merged[key] = value2
    return merged



import os
import sys
import socket
import getpass
import pathlib
import datetime
from unittest.mock import patch
from typing import Optional, Generator
from contextlib import ExitStack

import jwt
import shutil
import pytest

from .mock_slurm import mock_slurm_clients


def pytest_addoption(parser):
    parser.addoption(
        "--slurm-root-directory",
        action="store",
        default=None,
        help="Specify the SLURM root directory for logs and data.",
    )
    parser.addoption(
        "--slurm-api-version",
        action="store",
        default=None,
        help="Specify the SLURM API version.",
    )
    parser.addoption(
        "--slurm-pre-script",
        action="store",
        default=None,
        help="Execute at the start of the SLURM job.",
    )


@pytest.fixture(scope="session")
def slurm_root_directory(tmp_path_factory, request) -> pathlib.Path:
    path = request.config.getoption("--slurm-root-directory")
    if path:
        return pathlib.Path(path)
    return tmp_path_factory.mktemp("slurm_root")


@pytest.fixture(scope="session", params=["mock", "production"])
def slurm_parameters(request, tmp_path_factory):
    api_version = request.config.getoption("--slurm-api-version")
    pre_script = request.config.getoption("--slurm-pre-script")

    if request.param == "mock":
        with _patch_mock_defaults(tmp_path_factory):
            params = _get_mock_parameters()
            params.update(api_version=api_version, pre_script=pre_script)
            yield params

    elif request.param == "production":
        params = _get_production_parameters(request)
        params.update(api_version=api_version, pre_script=pre_script)
        yield params

    else:
        raise ValueError(request.param)


def _patch_mock_defaults(tmp_path_factory):
    from pyslurmutils.client import defaults

    stack = ExitStack()
    tmp_dir = tmp_path_factory.mktemp("slurm_tempdir")
    stack.enter_context(patch.object(defaults, "SLURM_TEMPDIR", str(tmp_dir)))

    if sys.platform == "win32":
        from pyslurmutils.client.rest import pyconn

        stack.enter_context(
            patch.object(pyconn, "_create_script", pyconn._create_windows_script)
        )
        stack.enter_context(patch.object(defaults, "SHEBANG", "#!powershell"))
        stack.enter_context(patch.object(defaults, "PYTHON_CMD", "python"))

    return stack


def _get_mock_parameters():
    return {
        "mock": True,
        "url": "http://mockhost",
        "user_name": "mockuser",
        "token": _create_mock_token("mockuser"),
        "api_version": None,  # will be overwritten by the fixture
        "pre_script": None,  # will be overwritten by the fixture
    }


def _get_production_parameters(request):
    url = os.environ.get("SLURM_URL")
    if not url:
        pytest.skip("SLURM_URL environment variable required")

    user_name = os.environ.get("SLURM_USER") or getpass.getuser()
    token = os.environ.get("SLURM_TOKEN")
    renewal_url = os.environ.get("SLURM_RENEWAL_URL")
    if not token and not renewal_url:
        pytest.skip("SLURM_TOKEN or SLURM_RENEWAL_URL environment variable required")

    slurm_root_directory = request.config.getoption("--slurm-root-directory")
    if not slurm_root_directory:
        pytest.skip("--slurm-root-directory pytest argument is required")

    return {
        "mock": False,
        "url": url,
        "user_name": user_name,
        "token": token,
        "renewal_url": renewal_url,
        "api_version": None,  # will be overwritten by the fixture
        "pre_script": None,  # will be overwritten by the fixture
    }


def _create_mock_token(username: str = "mockuser", exp_seconds: int = 3600) -> str:
    now_utc = datetime.datetime.now(datetime.timezone.utc)
    payload = {
        "sub": username,
        "exp": now_utc + datetime.timedelta(seconds=exp_seconds),
        "iat": now_utc,
    }
    secret = (
        "testsecret"  # Not used during decode (since signature verification is off)
    )
    return jwt.encode(payload, secret, algorithm="HS256")


@pytest.fixture(scope="session")
def slurm_log_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> pathlib.Path:
    if slurm_parameters["mock"]:
        return tmp_path_factory.mktemp("slurm_logs")
    return slurm_root_directory / slurm_parameters["user_name"] / "slurm_logs"


@pytest.fixture(scope="session")
def slurm_data_directory(
    slurm_root_directory, tmp_path_factory, slurm_parameters
) -> Optional[pathlib.Path]:
    if slurm_parameters["mock"]:
        return tmp_path_factory.mktemp("slurm_data")
    return slurm_root_directory / slurm_parameters["user_name"] / "slurm_data"


@pytest.fixture()
def slurm_tmp_path(
    slurm_root_directory, tmp_path, slurm_parameters
) -> Generator[pathlib.Path, None, None]:
    if slurm_parameters["mock"]:
        yield tmp_path
    else:
        tmp_path = (
            pathlib.Path(slurm_root_directory)
            / slurm_parameters["user_name"]
            / "pyslurmutils"
            / tmp_path.parent.stem
            / tmp_path.stem
        )
        tmp_path.mkdir(parents=True, exist_ok=True)
        yield tmp_path
        shutil.rmtree(tmp_path.parent, ignore_errors=True)


@pytest.fixture(scope="session")
def mock_job_name():
    job_name = f"pyslurmutils.unittest.{socket.gethostname()}"
    with patch("pyslurmutils.client.defaults.JOB_NAME", job_name):
        yield job_name


@pytest.fixture(scope="session")
def slurm_client_kwargs(
    slurm_log_directory, tmp_path_factory, slurm_parameters, mock_job_name
) -> Generator[dict, None, None]:
    # This fixture must remain a session scope fixture for third-party libraries

    params = dict(slurm_parameters)
    _ = params.pop("mock")
    params["log_directory"] = slurm_log_directory
    if slurm_parameters["mock"]:
        tmp_path = tmp_path_factory.mktemp("slurm_mock")
        with mock_slurm_clients(tmp_path, slurm_parameters):
            yield params
    else:
        yield params



import logging
from typing import Any, Optional, Tuple

from ._connection_base import Connection
from ._connection_file import FileConnection
from ._connection_tcp import TcpConnection
from ._proxy_base import WorkerProxy
from ..remote import remote_script

logger = logging.getLogger(__name__)


class RemoteWorkerProxy(WorkerProxy):
    """Worker proxy which executes tasks remotely over a :code:`Connection`."""

    def __init__(
        self,
        connection: Connection,
        initializer: Optional[callable] = None,
        initargs: Optional[tuple] = None,
        initkwargs: Optional[tuple] = None,
        max_tasks: Optional[int] = None,
    ) -> None:
        self._connection = connection

        if isinstance(connection, FileConnection):
            self._remote_environment = {
                "_PYSLURMUTILS_INFILE": connection.input_filename,
                "_PYSLURMUTILS_OUTFILE": connection.output_filename,
            }
            self._metadata = {
                "connection": "file",
                "infile": connection.input_filename,
                "outfile": connection.output_filename,
            }
        elif isinstance(connection, TcpConnection):
            self._remote_environment = {
                "_PYSLURMUTILS_HOST": connection.host,
                "_PYSLURMUTILS_PORT": connection.port,
            }
            self._metadata = {"connection": "tcp"}
        else:
            raise TypeError(connection)

        super().__init__(
            initializer=initializer,
            initargs=initargs,
            initkwargs=initkwargs,
            max_tasks=max_tasks,
        )

    def remote_script(self) -> str:
        return remote_script(self._metadata["connection"])

    @property
    def remote_environment(self) -> dict:
        return self._remote_environment

    @property
    def metadata(self) -> dict:
        return self._metadata

    def _send(self, data: Any) -> None:
        self._connection.send_data(data)

    def _get_result(self) -> Tuple[Any, Optional[BaseException]]:
        return self._connection.receive_data()



import os
import time
import logging
import threading
from typing import Generator, List, Optional
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor


def monitor_log_file(
    file_path: str, stop_event: threading.Event, period: float = 0.5
) -> None:
    """Redirect a log file with the following structure to the local root logger

    .. code

        [SLURM15733578] [INFO] [root] Your log message here
        [SLURM15733578] [ERROR] [package.module] Your log message here
        ...
    """
    last_log_level = logging.INFO
    current_record = list()

    dir_name = os.path.dirname(file_path)
    while not os.path.exists(file_path) and not stop_event.is_set():
        time.sleep(period)
        _ = os.listdir(dir_name)

    if stop_event.is_set():
        return

    with open(file_path, "r") as log_file:
        while not stop_event.is_set():
            line = log_file.readline()
            if line:
                try:
                    parts = line.split(" ", 3)
                    job_id = parts[0].strip("[]")
                    level = parts[1].strip("[]")
                    name = parts[2].strip("[]")
                    message = parts[3].rstrip()

                    log_level = getattr(logging, level)
                    if not isinstance(log_level, int):
                        raise TypeError

                    if current_record:
                        logging.log(last_log_level, "\n".join(current_record))
                        current_record.clear()
                    current_record.append(f"[{job_id}] [{name}] {message}")
                    last_log_level = log_level
                except (IndexError, AttributeError, TypeError):
                    current_record.append(line.rstrip())
            else:
                time.sleep(period)


@contextmanager
def log_file_monitor_context(
    file_paths: List[Optional[str]],
) -> Generator[None, None, None]:
    """Redirect logs files to the local root logger within this context."""
    file_paths = {
        file_path for file_path in file_paths if file_path and file_path != "/dev/null"
    }
    if not file_paths:
        yield
        return

    stop_event = threading.Event()
    with ThreadPoolExecutor(max_workers=len(file_paths)) as executor:
        futures = [
            executor.submit(monitor_log_file, file_path, stop_event)
            for file_path in file_paths
        ]
        try:
            yield
        finally:
            stop_event.set()
            for future, file_path in zip(futures, file_paths):
                try:
                    future.result()
                except Exception:
                    logging.warning("monitoring %s failed", file_path, exc_info=True)



import os
import shlex
import getpass
import subprocess
import datetime
import logging
from urllib.parse import urlparse, ParseResult
from typing import Optional, List

import jwt

from ... import url_utils
from ...errors import SlurmMissingParameterError
from ...errors import SlurmTokenInvalidError
from ...errors import SlurmTokenRenewalError
from ...errors import SlurmInvalidUrlError

logger = logging.getLogger(__name__)


class SlurmAuthManager:
    def __init__(
        self,
        user_name: Optional[str] = None,
        token: Optional[str] = None,
        renewal_url: Optional[str] = None,
        renewal_margin_seconds: int = 60,
        renewal_lifespan: int = 3600,
        renewal_timeout: int = 6,
        use_os_environment: bool = True,
    ):
        """
        Initializes the SLURM authentication manager.

        :param user_name: SLURM username (fallback to SLURM_USER or system user)
        :param token: SLURM JWT token (fallback to SLURM_TOKEN env)
        :param renewal_url: Url for SLURM JWT token renewal (fallback to SLURM_RENEWAL_URL env)
        :param renewal_margin_seconds: Time before expiry to trigger renewal
        :param renewal_lifespan: Requested lifespan in seconds for a new token
        :param renewal_timeout: Token renewal timeout in seconds
        :param use_os_environment: Use ``SLURM_*`` environment variables
        :raises SlurmTokenInvalidError:
        :raises SlurmInvalidUrlError:
        """
        self._use_os_environment = use_os_environment

        if not user_name and use_os_environment:
            user_name = os.environ.get("SLURM_USER")
        if not user_name:
            user_name = getpass.getuser()
        self._user_name = user_name.strip()

        if not token and use_os_environment:
            token = os.environ.get("SLURM_TOKEN")
        self._token = token

        if not renewal_url and use_os_environment:
            renewal_url = os.environ.get("SLURM_RENEWAL_URL")
        if renewal_url:
            self._renewal_url = _parse_renewal_url(renewal_url)
        else:
            self._renewal_url = None

        self._renewal_lifespan = renewal_lifespan
        self._renewal_margin_seconds = datetime.timedelta(
            seconds=renewal_margin_seconds
        )
        self._renewal_timeout = renewal_timeout

    @property
    def user_name(self) -> str:
        return self._user_name

    @property
    def _token(self) -> Optional[str]:
        return self.__token

    @_token.setter
    def _token(self, value: Optional[str]) -> None:
        """
        :raises SlurmTokenInvalidError:
        """
        if value is not None and not isinstance(value, str):
            raise TypeError(f"Token must be str or None, got {type(value).__name__}")
        if value:
            value = _parse_token(value)
            self.__token = value
            if self._use_os_environment:
                os.environ["SLURM_TOKEN"] = value
        else:
            self.__token = None
            if self._use_os_environment:
                _ = os.environ.pop("SLURM_TOKEN", None)

    def headers(self) -> dict:
        """
        Returns authentication headers for a SLURM request.

        If the current token is expired, a new token is retrieved automatically.

        :return: Dictionary containing authentication headers
        :raises SlurmTokenRenewalError:
        :raises SlurmTokenInvalidError:
        :raises SlurmMissingParameterError:
        """
        renewal_required = self._token_renewal_required()
        if renewal_required:
            self._token = self._create_new_token()
        self._raise_on_missing_token()
        return {
            "X-SLURM-USER-NAME": self._user_name,
            "X-SLURM-USER-TOKEN": self._token,
        }

    def retry_headers(self) -> Optional[dict]:
        """
        Returns authentication headers for a retry attempt on a SLURM request.

        Triggers token renewal if the token is close to expiring (within the configured margin).

        :return: Dictionary containing authentication headers, or None if the token is not renewed
        :raises SlurmTokenRenewalError:
        :raises SlurmTokenInvalidError:
        :raises SlurmMissingParameterError:
        """
        renewal_required = self._token_renewal_required(use_offset=True)
        if renewal_required:
            self._token = self._create_new_token()
            self._raise_on_missing_token()
            return {
                "X-SLURM-USER-NAME": self._user_name,
                "X-SLURM-USER-TOKEN": self._token,
            }
        return None

    def _token_renewal_required(self, use_offset: bool = False) -> bool:
        """
        Determines whether the current token should be renewed.

        :param use_offset: Use configured renewal_margin_seconds for early renewal
        :return: True if token should be renewed
        """
        if not self._token:
            return True

        try:
            payload = jwt.decode(self._token, options={"verify_signature": False})
        except jwt.DecodeError:
            return True

        exp = payload.get("exp")
        if not exp:
            return True

        expiry_time = datetime.datetime.fromtimestamp(exp).astimezone(
            datetime.timezone.utc
        )
        margin = self._renewal_margin_seconds if use_offset else datetime.timedelta(0)
        time_left = expiry_time - datetime.datetime.now(datetime.timezone.utc)
        expired = time_left <= margin
        if expired:
            logger.warning("SLURM token expired on %s", expiry_time)
        return expired

    def _create_new_token(self) -> str:
        """
        Attempts to renew the SLURM token (local first, then remote).

        :return: New token as a string
        :raises SlurmTokenRenewalError:
        :raises SlurmTokenInvalidError:
        :raises SlurmMissingParameterError:
        """
        new_token = self._create_new_token_local()
        if new_token:
            return new_token
        return self._create_new_token_ssh()

    def _create_new_token_local(self) -> Optional[str]:
        """
        Attempts to renew the token using local `scontrol`.

        :return: New token if successful, else None
        :raises SlurmTokenInvalidError:
        """
        cmd = self._renewal_command()
        return self._execute_renewal(cmd)

    def _create_new_token_ssh(self) -> str:
        """
        Attempts to renew the token via SSH to the configured renewal host.

        :return: New token if successful
        :raises SlurmTokenRenewalError:
        :raises SlurmTokenInvalidError:
        :raises SlurmMissingParameterError:
        """
        if not self._renewal_url:
            self._raise_on_missing_token()
            raise SlurmMissingParameterError(
                "SLURM token renewal failed: 'renewal_url' most be provided or set environment variable SLURM_RENEWAL_URL."
            )
        cmd = self._renewal_command()
        user = self._renewal_url.username or self._user_name
        destination = f"{user}@{self._renewal_url.hostname}"
        ssh_cmd = ["ssh", destination, "-p", str(self._renewal_url.port), " ".join(cmd)]
        return self._execute_renewal(
            ssh_cmd, timeout=self._renewal_timeout, raise_on_error=True
        )

    def _renewal_command(self) -> List[str]:
        """
        Constructs the `scontrol token` command.

        :return: List of command components
        """
        return [
            "scontrol",
            "token",
            f"username={self._user_name}",
            f"lifespan={self._renewal_lifespan}",
        ]

    def _execute_renewal(
        self, cmd: List[str], raise_on_error: bool = False, **kwargs
    ) -> Optional[str]:
        """
        Executes a command to retrieve a new SLURM token.

        :param cmd: Command as list of strings
        :param raise_on_error: Raise when failed
        :param kwargs: Additional subprocess parameters (e.g., timeout)
        :return: Parsed token if successful, else None
        :raises SlurmTokenInvalidError: If the renewal failed
        """
        s_cmd = shlex.join(cmd)
        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, universal_newlines=True, **kwargs
            )
            if output:
                logger.info("Successfully renewed SLURM token via: %s", s_cmd)
                return output
        except Exception as ex:
            if raise_on_error:
                self._raise_on_missing_token()
                raise SlurmTokenRenewalError(
                    f"SLURM token renewal failed: {s_cmd}"
                ) from ex
        return None

    def _raise_on_missing_token(self):
        """
        :raises SlurmMissingParameterError:
        """
        if not self._token:
            raise SlurmMissingParameterError(
                "SLURM authentication failed: 'token' must be provided or set environment variable SLURM_TOKEN or SLURM_RENEWAL_URL."
            )


def _parse_token(token: str) -> str:
    """
    Cleans and validates a SLURM JWT token.

    :param token: Raw token string
    :return: Cleaned token string
    :raises SlurmTokenInvalidError:
    """
    result = token.strip().split("\n")[-1]
    result = result.replace("SLURM_JWT=", "")
    if not result:
        raise SlurmTokenInvalidError("Slurm token is empty.")
    try:
        jwt.decode(result, options={"verify_signature": False})
    except jwt.DecodeError:
        raise SlurmTokenInvalidError("SLURM token is invalid.") from None
    return result


def _parse_renewal_url(renewal_url: str, default_port: int = 22) -> ParseResult:
    """
    :raises SlurmInvalidUrlError:
    """
    parsed = urlparse(renewal_url)
    if parsed.scheme != "ssh":
        raise SlurmInvalidUrlError(
            f"Invalid SLURM renewal URL scheme: {parsed.scheme}. Expected 'ssh'."
        )

    if not parsed.hostname:
        raise SlurmInvalidUrlError("Missing hostname.")

    if parsed.port is None:
        return url_utils.set_url_port(parsed, default_port)

    return parsed



# This file is automatically generated.
# DO NOT MODIFY THIS FILE MANUALLY.

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class DeleteBodySlurmJobJobid(BaseModel):
    pass


class DeletePathSlurmJobJobid(BaseModel):
    job_id: str = Field(..., title="Job Id")


class DeleteQuerySlurmJobJobid(BaseModel):
    signal: Optional[str] = Field(None, title="Signal")


class GetBodySlurmJobJobid(BaseModel):
    pass


class GetBodySlurmJobs(BaseModel):
    pass


class GetPathSlurmJobJobid(BaseModel):
    job_id: str = Field(..., title="Job Id")


class GetPathSlurmJobs(BaseModel):
    pass


class GetQuerySlurmJobJobid(BaseModel):
    update_time: Optional[str] = Field(None, title="Update Time")


class GetQuerySlurmJobs(BaseModel):
    update_time: Optional[str] = Field(None, title="Update Time")


class JobDescMsgBeginTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgCrontabLine(BaseModel):
    start: Optional[int] = Field(None, title="Start of this entry in file")
    end: Optional[int] = Field(None, title="End of this entry in file")


class JobDescMsgDistributionPlaneSize(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgKillWarningDelay(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgMemoryPerCpu(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgMemoryPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgPriority(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRequiredSwitches(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsAs(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsCore(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsCpu(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsData(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsFsize(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsMemlock(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsNofile(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsNproc(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsRss(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgRlimitsStack(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgSegmentSize(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgTimeLimit(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class JobDescMsgTimeMinimum(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespErrorsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form error description")
    error_number: Optional[int] = Field(None, title="Slurm numeric error identifier")
    error: Optional[str] = Field(None, title="Short form error description")
    source: Optional[str] = Field(
        None, title="Source of error or where error was first detected"
    )


class OpenapiJobInfoRespJobsItemAccrueTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemArrayJobId(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemArrayMaxTasks(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemArrayTaskId(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemBillableTres(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[float] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCoresPerSocket(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCpuFrequencyGovernor(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCpuFrequencyMaximum(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCpuFrequencyMinimum(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCpus(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemCpusPerTask(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemDeadline(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemDelayBoot(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemDerivedExitCodeReturnCode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemDerivedExitCodeSignalId(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemEligibleTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemEndTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemExitCodeReturnCode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemExitCodeSignalId(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemHetJobId(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemHetJobOffset(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemCpus(BaseModel):
    count: Optional[int] = Field(None, title="Total number of CPUs assigned to job")
    used: Optional[int] = Field(None, title="Total number of CPUs used by job")


class OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemMemory(BaseModel):
    used: Optional[int] = Field(None, title="Total memory (MiB) used by job")
    allocated: Optional[int] = Field(None, title="Total memory (MiB) allocated to job")


class OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItemCoresItem(
    BaseModel
):
    index: Optional[int] = Field(None, title="Core index")
    status: Optional[List[str]] = Field(None, title="Core status")


class OpenapiJobInfoRespJobsItemJobResourcesThreadsPerCore(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemLastSchedEvaluation(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMaxCpus(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMaxNodes(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMemoryPerCpu(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMemoryPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMinimumCpusPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemMinimumTmpDiskPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemNodeCount(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemPower(BaseModel):
    flags: Optional[List[Dict[str, Any]]] = Field(None, title="Flags")


class OpenapiJobInfoRespJobsItemPreSusTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemPreemptTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemPreemptableTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemPriority(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemResizeTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemSocketsPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemStartTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemSubmitTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemSuspendTime(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasks(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasksPerBoard(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasksPerCore(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasksPerNode(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasksPerSocket(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTasksPerTres(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemThreadsPerCore(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTimeLimit(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespJobsItemTimeMinimum(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespLastBackfill(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespLastUpdate(BaseModel):
    set: Optional[bool] = Field(
        None, title="True if number has been set; False if number is unset"
    )
    infinite: Optional[bool] = Field(
        None,
        title='True if number has been set to infinite; "set" and "number" will be ignored',
    )
    number: Optional[int] = Field(
        None,
        title='If "set" is True the number will be set with value; otherwise ignore number contents',
    )


class OpenapiJobInfoRespMetaClient(BaseModel):
    source: Optional[str] = Field(None, title="Client source description")
    user: Optional[str] = Field(None, title="Client user (if known)")
    group: Optional[str] = Field(None, title="Client group (if known)")


class OpenapiJobInfoRespMetaPlugin(BaseModel):
    type: Optional[str] = Field(None, title="Slurm plugin type (if applicable)")
    name: Optional[str] = Field(None, title="Slurm plugin name (if applicable)")
    data_parser: Optional[str] = Field(None, title="Slurm data_parser plugin")
    accounting_storage: Optional[str] = Field(None, title="Slurm accounting plugin")


class OpenapiJobInfoRespMetaSlurmVersion(BaseModel):
    major: Optional[str] = Field(None, title="Slurm release major version")
    micro: Optional[str] = Field(None, title="Slurm release micro version")
    minor: Optional[str] = Field(None, title="Slurm release minor version")


class OpenapiJobInfoRespWarningsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form warning description")
    source: Optional[str] = Field(
        None, title="Source of warning or where warning was first detected"
    )


class OpenapiJobSubmitResponseErrorsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form error description")
    error_number: Optional[int] = Field(None, title="Slurm numeric error identifier")
    error: Optional[str] = Field(None, title="Short form error description")
    source: Optional[str] = Field(
        None, title="Source of error or where error was first detected"
    )


class OpenapiJobSubmitResponseMetaClient(BaseModel):
    source: Optional[str] = Field(None, title="Client source description")
    user: Optional[str] = Field(None, title="Client user (if known)")
    group: Optional[str] = Field(None, title="Client group (if known)")


class OpenapiJobSubmitResponseMetaPlugin(BaseModel):
    type: Optional[str] = Field(None, title="Slurm plugin type (if applicable)")
    name: Optional[str] = Field(None, title="Slurm plugin name (if applicable)")
    data_parser: Optional[str] = Field(None, title="Slurm data_parser plugin")
    accounting_storage: Optional[str] = Field(None, title="Slurm accounting plugin")


class OpenapiJobSubmitResponseMetaSlurmVersion(BaseModel):
    major: Optional[str] = Field(None, title="Slurm release major version")
    micro: Optional[str] = Field(None, title="Slurm release micro version")
    minor: Optional[str] = Field(None, title="Slurm release minor version")


class OpenapiJobSubmitResponseResult(BaseModel):
    job_id: Optional[int] = Field(None, title="New job ID")
    step_id: Optional[str] = Field(None, title="New job step ID")
    error_code: Optional[int] = Field(None, title="Error code")
    error: Optional[str] = Field(None, title="Error message")
    job_submit_user_msg: Optional[str] = Field(
        None, title="Message to user from job_submit plugin"
    )


class OpenapiJobSubmitResponseWarningsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form warning description")
    source: Optional[str] = Field(
        None, title="Source of warning or where warning was first detected"
    )


class OpenapiRespErrorsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form error description")
    error_number: Optional[int] = Field(None, title="Slurm numeric error identifier")
    error: Optional[str] = Field(None, title="Short form error description")
    source: Optional[str] = Field(
        None, title="Source of error or where error was first detected"
    )


class OpenapiRespMetaClient(BaseModel):
    source: Optional[str] = Field(None, title="Client source description")
    user: Optional[str] = Field(None, title="Client user (if known)")
    group: Optional[str] = Field(None, title="Client group (if known)")


class OpenapiRespMetaPlugin(BaseModel):
    type: Optional[str] = Field(None, title="Slurm plugin type (if applicable)")
    name: Optional[str] = Field(None, title="Slurm plugin name (if applicable)")
    data_parser: Optional[str] = Field(None, title="Slurm data_parser plugin")
    accounting_storage: Optional[str] = Field(None, title="Slurm accounting plugin")


class OpenapiRespMetaSlurmVersion(BaseModel):
    major: Optional[str] = Field(None, title="Slurm release major version")
    micro: Optional[str] = Field(None, title="Slurm release micro version")
    minor: Optional[str] = Field(None, title="Slurm release minor version")


class OpenapiRespWarningsItem(BaseModel):
    description: Optional[str] = Field(None, title="Long form warning description")
    source: Optional[str] = Field(
        None, title="Source of warning or where warning was first detected"
    )


class PostPathSlurmJobSubmit(BaseModel):
    pass


class PostQuerySlurmJobSubmit(BaseModel):
    pass


class JobDescMsgCrontab(BaseModel):
    flags: Optional[List[str]] = Field(None, title="Flags")
    minute: Optional[str] = Field(
        None, title="Ranged string specifying eligible minute values (e.g. 0-10,50)"
    )
    hour: Optional[str] = Field(
        None, title="Ranged string specifying eligible hour values (e.g. 0-5,23)"
    )
    day_of_month: Optional[str] = Field(
        None,
        title="Ranged string specifying eligible day of month values (e.g. 0-10,29)",
    )
    month: Optional[str] = Field(
        None, title="Ranged string specifying eligible month values (e.g. 0-5,12)"
    )
    day_of_week: Optional[str] = Field(
        None, title="Ranged string specifying eligible day of week values (e.g.0-3,7)"
    )
    specification: Optional[str] = Field(
        None,
        title="Time specification (* means valid for all allowed values) - minute hour day_of_month month day_of_week",
    )
    command: Optional[str] = Field(None, title="Command to run")
    line: Optional[JobDescMsgCrontabLine] = Field(
        None, title="Job desc msg crontab line"
    )


class JobDescMsgRlimits(BaseModel):
    cpu: Optional[JobDescMsgRlimitsCpu] = Field(
        None, title="Per-process CPU limit, in seconds."
    )
    fsize: Optional[JobDescMsgRlimitsFsize] = Field(
        None, title="Largest file that can be created, in bytes."
    )
    data: Optional[JobDescMsgRlimitsData] = Field(
        None, title="Maximum size of data segment, in bytes. "
    )
    stack: Optional[JobDescMsgRlimitsStack] = Field(
        None, title="Maximum size of stack segment, in bytes."
    )
    core: Optional[JobDescMsgRlimitsCore] = Field(
        None, title="Largest core file that can be created, in bytes."
    )
    rss: Optional[JobDescMsgRlimitsRss] = Field(
        None,
        title="Largest resident set size, in bytes. This affects swapping; processes that are exceeding their resident set size will be more likely to have physical memory taken from them.",
    )
    nproc: Optional[JobDescMsgRlimitsNproc] = Field(None, title="Number of processes.")
    nofile: Optional[JobDescMsgRlimitsNofile] = Field(
        None, title="Number of open files."
    )
    memlock: Optional[JobDescMsgRlimitsMemlock] = Field(
        None, title="Locked-in-memory address space"
    )
    as_: Optional[JobDescMsgRlimitsAs] = Field(
        None, alias="as", title="Address space limit."
    )


class OpenapiJobInfoRespJobsItemDerivedExitCodeSignal(BaseModel):
    id: Optional[OpenapiJobInfoRespJobsItemDerivedExitCodeSignalId] = Field(
        None, title="Signal sent to process (numeric)"
    )
    name: Optional[str] = Field(None, title="Signal sent to process")


class OpenapiJobInfoRespJobsItemExitCodeSignal(BaseModel):
    id: Optional[OpenapiJobInfoRespJobsItemExitCodeSignalId] = Field(
        None, title="Signal sent to process (numeric)"
    )
    name: Optional[str] = Field(None, title="Signal sent to process")


class OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItem(BaseModel):
    index: Optional[int] = Field(None, title="Core index")
    cores: Optional[
        List[
            OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItemCoresItem
        ]
    ] = Field(None, title="Core in socket")


class OpenapiJobInfoRespMetaSlurm(BaseModel):
    version: Optional[OpenapiJobInfoRespMetaSlurmVersion] = Field(
        None, title="Openapi job info resp meta slurm version"
    )
    release: Optional[str] = Field(None, title="Slurm release string")
    cluster: Optional[str] = Field(None, title="Slurm cluster name")


class OpenapiJobSubmitResponseMetaSlurm(BaseModel):
    version: Optional[OpenapiJobSubmitResponseMetaSlurmVersion] = Field(
        None, title="Openapi job submit response meta slurm version"
    )
    release: Optional[str] = Field(None, title="Slurm release string")
    cluster: Optional[str] = Field(None, title="Slurm cluster name")


class OpenapiRespMetaSlurm(BaseModel):
    version: Optional[OpenapiRespMetaSlurmVersion] = Field(
        None, title="Openapi resp meta slurm version"
    )
    release: Optional[str] = Field(None, title="Slurm release string")
    cluster: Optional[str] = Field(None, title="Slurm cluster name")


class JobDescMsg(BaseModel):
    account: Optional[str] = Field(None, title="Account associated with the job")
    account_gather_frequency: Optional[str] = Field(
        None, title="Job accounting and profiling sampling intervals in seconds"
    )
    admin_comment: Optional[str] = Field(
        None, title="Arbitrary comment made by administrator"
    )
    allocation_node_list: Optional[str] = Field(
        None, title="Local node making the resource allocation"
    )
    allocation_node_port: Optional[int] = Field(
        None, title="Port to send allocation confirmation to"
    )
    argv: Optional[List[str]] = Field(None, title="Arguments to the script")
    array: Optional[str] = Field(None, title="Job array index value specification")
    batch_features: Optional[str] = Field(
        None, title="Features required for batch script's node"
    )
    begin_time: Optional[JobDescMsgBeginTime] = Field(
        None,
        title="Defer the allocation of the job until the specified time (UNIX timestamp)",
    )
    flags: Optional[List[str]] = Field(None, title="Job flags")
    burst_buffer: Optional[str] = Field(None, title="Burst buffer specifications")
    clusters: Optional[str] = Field(
        None, title="Clusters that a federated job can run on"
    )
    cluster_constraint: Optional[str] = Field(
        None,
        title="Required features that a federated cluster must have to have a sibling job submitted to it",
    )
    comment: Optional[str] = Field(None, title="Arbitrary comment made by user")
    contiguous: Optional[bool] = Field(
        None, title="True if job requires contiguous nodes"
    )
    container: Optional[str] = Field(
        None, title="Absolute path to OCI container bundle"
    )
    container_id: Optional[str] = Field(None, title="OCI container ID")
    core_specification: Optional[int] = Field(None, title="Specialized core count")
    thread_specification: Optional[int] = Field(None, title="Specialized thread count")
    cpu_binding: Optional[str] = Field(
        None, title="Method for binding tasks to allocated CPUs"
    )
    cpu_binding_flags: Optional[List[str]] = Field(None, title="Flags for CPU binding")
    cpu_frequency: Optional[str] = Field(
        None, title="Requested CPU frequency range <p1>[-p2][:p3]"
    )
    cpus_per_tres: Optional[str] = Field(
        None,
        title="Semicolon delimited list of TRES=# values values indicating how many CPUs should be allocated for each specified TRES (currently only used for gres/gpu)",
    )
    crontab: Optional[JobDescMsgCrontab] = Field(
        None, title="Specification for scrontab job"
    )
    deadline: Optional[int] = Field(
        None, title="Latest time that the job may start (UNIX timestamp)"
    )
    delay_boot: Optional[int] = Field(
        None,
        title="Number of seconds after job eligible start that nodes will be rebooted to satisfy feature specification",
    )
    dependency: Optional[str] = Field(
        None,
        title="Other jobs that must meet certain criteria before this job can start",
    )
    end_time: Optional[int] = Field(None, title="Expected end time (UNIX timestamp)")
    environment: Optional[List[str]] = Field(
        None, title="Environment variables to be set for the job"
    )
    rlimits: Optional[JobDescMsgRlimits] = Field(None, title="Job desc msg rlimits")
    excluded_nodes: Optional[List[str]] = Field(
        None, title="Comma separated list of nodes that may not be used"
    )
    extra: Optional[str] = Field(
        None,
        title="Arbitrary string used for node filtering if extra constraints are enabled",
    )
    constraints: Optional[str] = Field(
        None, title="Comma separated list of features that are required"
    )
    group_id: Optional[str] = Field(
        None, title="Group ID of the user that owns the job"
    )
    hetjob_group: Optional[int] = Field(
        None,
        title="Unique sequence number applied to this component of the heterogeneous job",
    )
    immediate: Optional[bool] = Field(
        None,
        title="If true, exit if resources are not available within the time period specified",
    )
    job_id: Optional[int] = Field(None, title="Job ID")
    kill_on_node_fail: Optional[bool] = Field(
        None, title="If true, kill job on node failure"
    )
    licenses: Optional[str] = Field(None, title="License(s) required by the job")
    mail_type: Optional[List[str]] = Field(None, title="Mail event type(s)")
    mail_user: Optional[str] = Field(None, title="User to receive email notifications")
    mcs_label: Optional[str] = Field(
        None, title="Multi-Category Security label on the job"
    )
    memory_binding: Optional[str] = Field(None, title="Binding map for map/mask_cpu")
    memory_binding_type: Optional[List[str]] = Field(
        None, title="Method for binding tasks to memory"
    )
    memory_per_tres: Optional[str] = Field(
        None,
        title="Semicolon delimited list of TRES=# values indicating how much memory in megabytes should be allocated for each specified TRES (currently only used for gres/gpu)",
    )
    name: Optional[str] = Field(None, title="Job name")
    network: Optional[str] = Field(None, title="Network specs for job step")
    nice: Optional[int] = Field(None, title="Requested job priority change")
    tasks: Optional[int] = Field(None, title="Number of tasks")
    open_mode: Optional[List[str]] = Field(
        None, title="Open mode used for stdout and stderr files"
    )
    reserve_ports: Optional[int] = Field(
        None, title="Port to send various notification msg to"
    )
    overcommit: Optional[bool] = Field(None, title="Overcommit resources")
    partition: Optional[str] = Field(None, title="Partition assigned to the job")
    distribution_plane_size: Optional[JobDescMsgDistributionPlaneSize] = Field(
        None, title="Plane size specification when distribution specifies plane"
    )
    power_flags: Optional[List[Dict[str, Any]]] = Field(None, title="Power Flags")
    prefer: Optional[str] = Field(
        None,
        title="Comma separated list of features that are preferred but not required",
    )
    hold: Optional[bool] = Field(None, title="Hold (true) or release (false) job")
    priority: Optional[JobDescMsgPriority] = Field(
        None, title="Request specific job priority"
    )
    profile: Optional[List[str]] = Field(
        None, title="Profile used by the acct_gather_profile plugin"
    )
    qos: Optional[str] = Field(None, title="Quality of Service assigned to the job")
    reboot: Optional[bool] = Field(None, title="Node reboot requested before start")
    required_nodes: Optional[List[str]] = Field(
        None, title="Comma separated list of required nodes"
    )
    requeue: Optional[bool] = Field(
        None, title="Determines whether the job may be requeued"
    )
    reservation: Optional[str] = Field(None, title="Name of reservation to use")
    resv_mpi_ports: Optional[int] = Field(
        None,
        title="Number of reserved communication ports; can only be used if slurmstepd step manager is enabled",
    )
    script: Optional[str] = Field(
        None,
        title="Job batch script; only the first component in a HetJob is populated or honored",
    )
    shared: Optional[List[str]] = Field(
        None, title="How the job can share resources with other jobs, if at all"
    )
    exclusive: Optional[List[str]] = Field(None, title="Exclusive")
    oversubscribe: Optional[bool] = Field(None, title="Oversubscribe")
    site_factor: Optional[int] = Field(None, title="Site-specific priority factor")
    spank_environment: Optional[List[str]] = Field(
        None,
        title="Environment variables for job prolog/epilog scripts as set by SPANK plugins",
    )
    distribution: Optional[str] = Field(None, title="Layout")
    time_limit: Optional[JobDescMsgTimeLimit] = Field(
        None, title="Maximum run time in minutes"
    )
    time_minimum: Optional[JobDescMsgTimeMinimum] = Field(
        None, title="Minimum run time in minutes"
    )
    tres_bind: Optional[str] = Field(None, title="Task to TRES binding directives")
    tres_freq: Optional[str] = Field(None, title="TRES frequency directives")
    tres_per_job: Optional[str] = Field(
        None,
        title="Comma separated list of TRES=# values to be allocated for every job",
    )
    tres_per_node: Optional[str] = Field(
        None,
        title="Comma separated list of TRES=# values to be allocated for every node",
    )
    tres_per_socket: Optional[str] = Field(
        None,
        title="Comma separated list of TRES=# values to be allocated for every socket",
    )
    tres_per_task: Optional[str] = Field(
        None,
        title="Comma separated list of TRES=# values to be allocated for every task",
    )
    user_id: Optional[str] = Field(None, title="User ID that owns the job")
    wait_all_nodes: Optional[bool] = Field(
        None, title="If true, wait to start until after all nodes have booted"
    )
    kill_warning_flags: Optional[List[str]] = Field(
        None, title="Flags related to job signals"
    )
    kill_warning_signal: Optional[str] = Field(
        None, title='Signal to send when approaching end time (e.g. "10" or "USR1")'
    )
    kill_warning_delay: Optional[JobDescMsgKillWarningDelay] = Field(
        None, title="Number of seconds before end time to send the warning signal"
    )
    current_working_directory: Optional[str] = Field(
        None, title="Working directory to use for the job"
    )
    cpus_per_task: Optional[int] = Field(
        None, title="Number of CPUs required by each task"
    )
    minimum_cpus: Optional[int] = Field(None, title="Minimum number of CPUs required")
    maximum_cpus: Optional[int] = Field(None, title="Maximum number of CPUs required")
    nodes: Optional[str] = Field(
        None, title="Node count range specification (e.g. 1-15:4)"
    )
    minimum_nodes: Optional[int] = Field(None, title="Minimum node count")
    maximum_nodes: Optional[int] = Field(None, title="Maximum node count")
    minimum_boards_per_node: Optional[int] = Field(
        None, title="Boards per node required"
    )
    minimum_sockets_per_board: Optional[int] = Field(
        None, title="Sockets per board required"
    )
    sockets_per_node: Optional[int] = Field(None, title="Sockets per node required")
    threads_per_core: Optional[int] = Field(None, title="Threads per core required")
    tasks_per_node: Optional[int] = Field(
        None, title="Number of tasks to invoke on each node"
    )
    tasks_per_socket: Optional[int] = Field(
        None, title="Number of tasks to invoke on each socket"
    )
    tasks_per_core: Optional[int] = Field(
        None, title="Number of tasks to invoke on each core"
    )
    tasks_per_board: Optional[int] = Field(
        None, title="Number of tasks to invoke on each board"
    )
    ntasks_per_tres: Optional[int] = Field(
        None, title="Number of tasks that can access each GPU"
    )
    minimum_cpus_per_node: Optional[int] = Field(
        None, title="Minimum number of CPUs per node"
    )
    memory_per_cpu: Optional[JobDescMsgMemoryPerCpu] = Field(
        None, title="Minimum memory in megabytes per allocated CPU"
    )
    memory_per_node: Optional[JobDescMsgMemoryPerNode] = Field(
        None, title="Minimum memory in megabytes per allocated CPU"
    )
    temporary_disk_per_node: Optional[int] = Field(
        None, title="Minimum tmp disk space required per node"
    )
    selinux_context: Optional[str] = Field(None, title="SELinux context")
    required_switches: Optional[JobDescMsgRequiredSwitches] = Field(
        None, title="Maximum number of switches"
    )
    segment_size: Optional[JobDescMsgSegmentSize] = Field(
        None, title="Segment size for topology/block"
    )
    standard_error: Optional[str] = Field(None, title="Path to stderr file")
    standard_input: Optional[str] = Field(None, title="Path to stdin file")
    standard_output: Optional[str] = Field(None, title="Path to stdout file")
    wait_for_switch: Optional[int] = Field(
        None, title="Maximum time to wait for switches in seconds"
    )
    wckey: Optional[str] = Field(None, title="Workload characterization key")
    x11: Optional[List[str]] = Field(None, title="X11 forwarding options")
    x11_magic_cookie: Optional[str] = Field(
        None, title="Magic cookie for X11 forwarding"
    )
    x11_target_host: Optional[str] = Field(
        None, title="Hostname or UNIX socket if x11_target_port=0"
    )
    x11_target_port: Optional[int] = Field(None, title="TCP port")


class JobSubmitReq(BaseModel):
    script: Optional[str] = Field(
        None, title="Deprecated; Populate script field in jobs[0] or job"
    )
    jobs: Optional[List[JobDescMsg]] = Field(None, title="HetJob description")
    job: Optional[JobDescMsg] = Field(None, title="Job description")


class OpenapiJobInfoRespJobsItemDerivedExitCode(BaseModel):
    status: Optional[List[str]] = Field(None, title="Status given by return code")
    return_code: Optional[OpenapiJobInfoRespJobsItemDerivedExitCodeReturnCode] = Field(
        None, title="Process return code (numeric)"
    )
    signal: Optional[OpenapiJobInfoRespJobsItemDerivedExitCodeSignal] = Field(
        None, title="Openapi job info resp jobs item derived exit code signal"
    )


class OpenapiJobInfoRespJobsItemExitCode(BaseModel):
    status: Optional[List[str]] = Field(None, title="Status given by return code")
    return_code: Optional[OpenapiJobInfoRespJobsItemExitCodeReturnCode] = Field(
        None, title="Process return code (numeric)"
    )
    signal: Optional[OpenapiJobInfoRespJobsItemExitCodeSignal] = Field(
        None, title="Openapi job info resp jobs item exit code signal"
    )


class OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItem(BaseModel):
    index: Optional[int] = Field(None, title="Node index")
    name: Optional[str] = Field(None, title="Node name")
    cpus: Optional[OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemCpus] = (
        Field(
            None,
            title="Openapi job info resp jobs item job resources nodes allocation item cpus",
        )
    )
    memory: Optional[
        OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemMemory
    ] = Field(
        None,
        title="Openapi job info resp jobs item job resources nodes allocation item memory",
    )
    sockets: Optional[
        List[OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItem]
    ] = Field(None, title="Socket allocations in node")


class OpenapiJobInfoRespMeta(BaseModel):
    plugin: Optional[OpenapiJobInfoRespMetaPlugin] = Field(
        None, title="Openapi job info resp meta plugin"
    )
    client: Optional[OpenapiJobInfoRespMetaClient] = Field(
        None, title="Openapi job info resp meta client"
    )
    command: Optional[List[str]] = Field(None, title="CLI command (if applicable)")
    slurm: Optional[OpenapiJobInfoRespMetaSlurm] = Field(
        None, title="Openapi job info resp meta slurm"
    )


class OpenapiJobSubmitResponseMeta(BaseModel):
    plugin: Optional[OpenapiJobSubmitResponseMetaPlugin] = Field(
        None, title="Openapi job submit response meta plugin"
    )
    client: Optional[OpenapiJobSubmitResponseMetaClient] = Field(
        None, title="Openapi job submit response meta client"
    )
    command: Optional[List[str]] = Field(None, title="CLI command (if applicable)")
    slurm: Optional[OpenapiJobSubmitResponseMetaSlurm] = Field(
        None, title="Openapi job submit response meta slurm"
    )


class OpenapiRespMeta(BaseModel):
    plugin: Optional[OpenapiRespMetaPlugin] = Field(
        None, title="Openapi resp meta plugin"
    )
    client: Optional[OpenapiRespMetaClient] = Field(
        None, title="Openapi resp meta client"
    )
    command: Optional[List[str]] = Field(None, title="CLI command (if applicable)")
    slurm: Optional[OpenapiRespMetaSlurm] = Field(None, title="Openapi resp meta slurm")


class PostBodySlurmJobSubmit(BaseModel):
    content: Optional[JobSubmitReq] = Field(None, title="Job submit req")


class OpenapiJobInfoRespJobsItemJobResourcesNodes(BaseModel):
    count: Optional[int] = Field(None, title="Number of allocated nodes")
    select_type: Optional[List[str]] = Field(
        None, title="Node scheduling selection method"
    )
    list: Optional[str] = Field(None, title="Node(s) allocated to the job")
    whole: Optional[bool] = Field(None, title="Whether whole nodes were allocated")
    allocation: Optional[
        List[OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItem]
    ] = Field(None, title="Allocated node resources")


class OpenapiJobSubmitResponse(BaseModel):
    result: Optional[OpenapiJobSubmitResponseResult] = Field(
        None, title="Job submission"
    )
    job_id: Optional[int] = Field(None, title="Submitted Job ID")
    step_id: Optional[str] = Field(None, title="Submitted Step ID")
    job_submit_user_msg: Optional[str] = Field(
        None, title="Job submission user message"
    )
    meta: Optional[OpenapiJobSubmitResponseMeta] = Field(
        None, title="Slurm meta values"
    )
    errors: Optional[List[OpenapiJobSubmitResponseErrorsItem]] = Field(
        None, title="Query errors"
    )
    warnings: Optional[List[OpenapiJobSubmitResponseWarningsItem]] = Field(
        None, title="Query warnings"
    )


class OpenapiResp(BaseModel):
    meta: Optional[OpenapiRespMeta] = Field(None, title="Slurm meta values")
    errors: Optional[List[OpenapiRespErrorsItem]] = Field(None, title="Query errors")
    warnings: Optional[List[OpenapiRespWarningsItem]] = Field(
        None, title="Query warnings"
    )


class PostResponse200SlurmJobSubmit(BaseModel):
    content: OpenapiJobSubmitResponse = Field(..., title="Openapi job submit response")


class PostResponseDefaultSlurmJobSubmit(BaseModel):
    content: OpenapiJobSubmitResponse = Field(..., title="Openapi job submit response")


class DeleteResponse200SlurmJobJobid(BaseModel):
    content: OpenapiResp = Field(..., title="Openapi resp")


class DeleteResponseDefaultSlurmJobJobid(BaseModel):
    content: OpenapiResp = Field(..., title="Openapi resp")


class OpenapiJobInfoRespJobsItemJobResources(BaseModel):
    select_type: Optional[List[str]] = Field(
        None, title="Scheduler consumable resource selection type"
    )
    nodes: Optional[OpenapiJobInfoRespJobsItemJobResourcesNodes] = Field(
        None, title="Openapi job info resp jobs item job resources nodes"
    )
    cpus: Optional[int] = Field(None, title="Number of allocated CPUs")
    threads_per_core: Optional[OpenapiJobInfoRespJobsItemJobResourcesThreadsPerCore] = (
        Field(None, title="Number of processor threads per CPU core")
    )


class OpenapiJobInfoRespJobsItem(BaseModel):
    account: Optional[str] = Field(None, title="Account associated with the job")
    accrue_time: Optional[OpenapiJobInfoRespJobsItemAccrueTime] = Field(
        None, title="When the job started accruing age priority (UNIX timestamp)"
    )
    admin_comment: Optional[str] = Field(
        None, title="Arbitrary comment made by administrator"
    )
    allocating_node: Optional[str] = Field(
        None, title="Local node making the resource allocation"
    )
    array_job_id: Optional[OpenapiJobInfoRespJobsItemArrayJobId] = Field(
        None, title="Job ID of job array, or 0 if N/A"
    )
    array_task_id: Optional[OpenapiJobInfoRespJobsItemArrayTaskId] = Field(
        None, title="Task ID of this task in job array"
    )
    array_max_tasks: Optional[OpenapiJobInfoRespJobsItemArrayMaxTasks] = Field(
        None,
        title="Maximum number of simultaneously running array tasks, 0 if no limit",
    )
    array_task_string: Optional[str] = Field(
        None, title="String expression of task IDs in this record"
    )
    association_id: Optional[int] = Field(
        None, title="Unique identifier for the association"
    )
    batch_features: Optional[str] = Field(
        None, title="Features required for batch script's node"
    )
    batch_flag: Optional[bool] = Field(None, title="True if batch job")
    batch_host: Optional[str] = Field(None, title="Name of host running batch script")
    flags: Optional[List[str]] = Field(None, title="Job flags")
    burst_buffer: Optional[str] = Field(None, title="Burst buffer specifications")
    burst_buffer_state: Optional[str] = Field(None, title="Burst buffer state details")
    cluster: Optional[str] = Field(None, title="Cluster name")
    cluster_features: Optional[str] = Field(
        None, title="List of required cluster features"
    )
    command: Optional[str] = Field(None, title="Executed command")
    comment: Optional[str] = Field(None, title="Arbitrary comment")
    container: Optional[str] = Field(
        None, title="Absolute path to OCI container bundle"
    )
    container_id: Optional[str] = Field(None, title="OCI container ID")
    contiguous: Optional[bool] = Field(
        None, title="True if job requires contiguous nodes"
    )
    core_spec: Optional[int] = Field(None, title="Specialized core count")
    thread_spec: Optional[int] = Field(None, title="Specialized thread count")
    cores_per_socket: Optional[OpenapiJobInfoRespJobsItemCoresPerSocket] = Field(
        None, title="Cores per socket required"
    )
    billable_tres: Optional[OpenapiJobInfoRespJobsItemBillableTres] = Field(
        None, title="Billable TRES"
    )
    cpus_per_task: Optional[OpenapiJobInfoRespJobsItemCpusPerTask] = Field(
        None, title="Number of CPUs required by each task"
    )
    cpu_frequency_minimum: Optional[OpenapiJobInfoRespJobsItemCpuFrequencyMinimum] = (
        Field(None, title="Minimum CPU frequency")
    )
    cpu_frequency_maximum: Optional[OpenapiJobInfoRespJobsItemCpuFrequencyMaximum] = (
        Field(None, title="Maximum CPU frequency")
    )
    cpu_frequency_governor: Optional[OpenapiJobInfoRespJobsItemCpuFrequencyGovernor] = (
        Field(None, title="CPU frequency governor")
    )
    cpus_per_tres: Optional[str] = Field(
        None,
        title="Semicolon delimited list of TRES=# values indicating how many CPUs should be allocated for each specified TRES (currently only used for gres/gpu)",
    )
    cron: Optional[str] = Field(None, title="Time specification for scrontab job")
    deadline: Optional[OpenapiJobInfoRespJobsItemDeadline] = Field(
        None, title="Latest time that the job may start (UNIX timestamp)"
    )
    delay_boot: Optional[OpenapiJobInfoRespJobsItemDelayBoot] = Field(
        None,
        title="Number of seconds after job eligible start that nodes will be rebooted to satisfy feature specification",
    )
    dependency: Optional[str] = Field(
        None,
        title="Other jobs that must meet certain criteria before this job can start",
    )
    derived_exit_code: Optional[OpenapiJobInfoRespJobsItemDerivedExitCode] = Field(
        None, title="Highest exit code of all job steps"
    )
    eligible_time: Optional[OpenapiJobInfoRespJobsItemEligibleTime] = Field(
        None, title="Time when the job became eligible to run (UNIX timestamp)"
    )
    end_time: Optional[OpenapiJobInfoRespJobsItemEndTime] = Field(
        None, title="End time, real or expected (UNIX timestamp)"
    )
    excluded_nodes: Optional[str] = Field(
        None, title="Comma separated list of nodes that may not be used"
    )
    exit_code: Optional[OpenapiJobInfoRespJobsItemExitCode] = Field(
        None, title="Exit code of the job"
    )
    extra: Optional[str] = Field(
        None,
        title="Arbitrary string used for node filtering if extra constraints are enabled",
    )
    failed_node: Optional[str] = Field(
        None, title="Name of node that caused job failure"
    )
    features: Optional[str] = Field(
        None, title="Comma separated list of features that are required"
    )
    federation_origin: Optional[str] = Field(
        None, title="Origin cluster's name (when using federation)"
    )
    federation_siblings_active: Optional[str] = Field(
        None, title="Active sibling job names"
    )
    federation_siblings_viable: Optional[str] = Field(
        None, title="Viable sibling job names"
    )
    gres_detail: Optional[List[str]] = Field(
        None, title="List of GRES index and counts allocated per node"
    )
    group_id: Optional[int] = Field(
        None, title="Group ID of the user that owns the job"
    )
    group_name: Optional[str] = Field(
        None, title="Group name of the user that owns the job"
    )
    het_job_id: Optional[OpenapiJobInfoRespJobsItemHetJobId] = Field(
        None, title="Heterogeneous job ID, if applicable"
    )
    het_job_id_set: Optional[str] = Field(
        None, title="Job ID range for all heterogeneous job components"
    )
    het_job_offset: Optional[OpenapiJobInfoRespJobsItemHetJobOffset] = Field(
        None,
        title="Unique sequence number applied to this component of the heterogeneous job",
    )
    job_id: Optional[int] = Field(None, title="Job ID")
    job_resources: Optional[OpenapiJobInfoRespJobsItemJobResources] = Field(
        None, title="Resources used by the job"
    )
    job_size_str: Optional[List[str]] = Field(
        None, title="Number of nodes (in a range) required for this job"
    )
    job_state: Optional[List[str]] = Field(None, title="Current state")
    last_sched_evaluation: Optional[OpenapiJobInfoRespJobsItemLastSchedEvaluation] = (
        Field(None, title="Last time job was evaluated for scheduling (UNIX timestamp)")
    )
    licenses: Optional[str] = Field(None, title="License(s) required by the job")
    mail_type: Optional[List[str]] = Field(None, title="Mail event type(s)")
    mail_user: Optional[str] = Field(None, title="User to receive email notifications")
    max_cpus: Optional[OpenapiJobInfoRespJobsItemMaxCpus] = Field(
        None, title="Maximum number of CPUs usable by the job"
    )
    max_nodes: Optional[OpenapiJobInfoRespJobsItemMaxNodes] = Field(
        None, title="Maximum number of nodes usable by the job"
    )
    mcs_label: Optional[str] = Field(
        None, title="Multi-Category Security label on the job"
    )
    memory_per_tres: Optional[str] = Field(
        None,
        title="Semicolon delimited list of TRES=# values indicating how much memory in megabytes should be allocated for each specified TRES (currently only used for gres/gpu)",
    )
    name: Optional[str] = Field(None, title="Job name")
    network: Optional[str] = Field(None, title="Network specs for the job")
    nodes: Optional[str] = Field(None, title="Node(s) allocated to the job")
    nice: Optional[int] = Field(None, title="Requested job priority change")
    tasks_per_core: Optional[OpenapiJobInfoRespJobsItemTasksPerCore] = Field(
        None, title="Number of tasks invoked on each core"
    )
    tasks_per_tres: Optional[OpenapiJobInfoRespJobsItemTasksPerTres] = Field(
        None, title="Number of tasks that can assess each GPU"
    )
    tasks_per_node: Optional[OpenapiJobInfoRespJobsItemTasksPerNode] = Field(
        None, title="Number of tasks invoked on each node"
    )
    tasks_per_socket: Optional[OpenapiJobInfoRespJobsItemTasksPerSocket] = Field(
        None, title="Number of tasks invoked on each socket"
    )
    tasks_per_board: Optional[OpenapiJobInfoRespJobsItemTasksPerBoard] = Field(
        None, title="Number of tasks invoked on each board"
    )
    cpus: Optional[OpenapiJobInfoRespJobsItemCpus] = Field(
        None, title="Minimum number of CPUs required"
    )
    node_count: Optional[OpenapiJobInfoRespJobsItemNodeCount] = Field(
        None, title="Minimum number of nodes required"
    )
    tasks: Optional[OpenapiJobInfoRespJobsItemTasks] = Field(
        None, title="Number of tasks"
    )
    partition: Optional[str] = Field(None, title="Partition assigned to the job")
    prefer: Optional[str] = Field(
        None, title="Feature(s) the job requested but that are not required"
    )
    memory_per_cpu: Optional[OpenapiJobInfoRespJobsItemMemoryPerCpu] = Field(
        None, title="Minimum memory in megabytes per allocated CPU"
    )
    memory_per_node: Optional[OpenapiJobInfoRespJobsItemMemoryPerNode] = Field(
        None, title="Minimum memory in megabytes per allocated node"
    )
    minimum_cpus_per_node: Optional[OpenapiJobInfoRespJobsItemMinimumCpusPerNode] = (
        Field(None, title="Minimum number of CPUs per node")
    )
    minimum_tmp_disk_per_node: Optional[
        OpenapiJobInfoRespJobsItemMinimumTmpDiskPerNode
    ] = Field(None, title="Minimum tmp disk space required per node")
    power: Optional[OpenapiJobInfoRespJobsItemPower] = Field(
        None, title="Openapi job info resp jobs item power"
    )
    preempt_time: Optional[OpenapiJobInfoRespJobsItemPreemptTime] = Field(
        None, title="Time job received preemption signal (UNIX timestamp)"
    )
    preemptable_time: Optional[OpenapiJobInfoRespJobsItemPreemptableTime] = Field(
        None, title="Time job becomes eligible for preemption (UNIX timestamp)"
    )
    pre_sus_time: Optional[OpenapiJobInfoRespJobsItemPreSusTime] = Field(
        None, title="Total run time prior to last suspend in seconds"
    )
    hold: Optional[bool] = Field(None, title="Hold (true) or release (false) job")
    priority: Optional[OpenapiJobInfoRespJobsItemPriority] = Field(
        None, title="Request specific job priority"
    )
    profile: Optional[List[str]] = Field(
        None, title="Profile used by the acct_gather_profile plugin"
    )
    qos: Optional[str] = Field(None, title="Quality of Service assigned to the job")
    reboot: Optional[bool] = Field(None, title="Node reboot requested before start")
    required_nodes: Optional[str] = Field(
        None, title="Comma separated list of required nodes"
    )
    minimum_switches: Optional[int] = Field(
        None, title="Maximum number of switches (the 'minimum' in the key is incorrect)"
    )
    requeue: Optional[bool] = Field(
        None, title="Determines whether the job may be requeued"
    )
    resize_time: Optional[OpenapiJobInfoRespJobsItemResizeTime] = Field(
        None, title="Time of last size change (UNIX timestamp)"
    )
    restart_cnt: Optional[int] = Field(None, title="Number of job restarts")
    resv_name: Optional[str] = Field(None, title="Name of reservation to use")
    scheduled_nodes: Optional[str] = Field(
        None, title="List of nodes scheduled to be used for the job"
    )
    selinux_context: Optional[str] = Field(None, title="SELinux context")
    shared: Optional[List[str]] = Field(
        None, title="How the job can share resources with other jobs, if at all"
    )
    exclusive: Optional[List[str]] = Field(None, title="Exclusive")
    oversubscribe: Optional[bool] = Field(None, title="Oversubscribe")
    show_flags: Optional[List[str]] = Field(
        None, title="Job details shown in this response"
    )
    sockets_per_board: Optional[int] = Field(
        None, title="Number of sockets per board required"
    )
    sockets_per_node: Optional[OpenapiJobInfoRespJobsItemSocketsPerNode] = Field(
        None, title="Number of sockets per node required"
    )
    start_time: Optional[OpenapiJobInfoRespJobsItemStartTime] = Field(
        None, title="Time execution began, or is expected to begin (UNIX timestamp)"
    )
    state_description: Optional[str] = Field(
        None, title="Optional details for state_reason"
    )
    state_reason: Optional[str] = Field(
        None, title="Reason for current Pending or Failed state"
    )
    standard_error: Optional[str] = Field(None, title="Path to stderr file")
    standard_input: Optional[str] = Field(None, title="Path to stdin file")
    standard_output: Optional[str] = Field(None, title="Path to stdout file")
    submit_time: Optional[OpenapiJobInfoRespJobsItemSubmitTime] = Field(
        None, title="Time when the job was submitted (UNIX timestamp)"
    )
    suspend_time: Optional[OpenapiJobInfoRespJobsItemSuspendTime] = Field(
        None, title="Time the job was last suspended or resumed (UNIX timestamp)"
    )
    system_comment: Optional[str] = Field(
        None, title="Arbitrary comment from slurmctld"
    )
    time_limit: Optional[OpenapiJobInfoRespJobsItemTimeLimit] = Field(
        None, title="Maximum run time in minutes"
    )
    time_minimum: Optional[OpenapiJobInfoRespJobsItemTimeMinimum] = Field(
        None, title="Minimum run time in minutes"
    )
    threads_per_core: Optional[OpenapiJobInfoRespJobsItemThreadsPerCore] = Field(
        None, title="Number of processor threads per CPU core required"
    )
    tres_bind: Optional[str] = Field(None, title="Task to TRES binding directives")
    tres_freq: Optional[str] = Field(None, title="TRES frequency directives")
    tres_per_job: Optional[str] = Field(
        None, title="Comma separated list of TRES=# values to be allocated per job"
    )
    tres_per_node: Optional[str] = Field(
        None, title="Comma separated list of TRES=# values to be allocated per node"
    )
    tres_per_socket: Optional[str] = Field(
        None, title="Comma separated list of TRES=# values to be allocated per socket"
    )
    tres_per_task: Optional[str] = Field(
        None, title="Comma separated list of TRES=# values to be allocated per task"
    )
    tres_req_str: Optional[str] = Field(None, title="TRES requested by the job")
    tres_alloc_str: Optional[str] = Field(None, title="TRES used by the job")
    user_id: Optional[int] = Field(None, title="User ID that owns the job")
    user_name: Optional[str] = Field(None, title="User name that owns the job")
    maximum_switch_wait_time: Optional[int] = Field(
        None, title="Maximum time to wait for switches in seconds"
    )
    wckey: Optional[str] = Field(None, title="Workload characterization key")
    current_working_directory: Optional[str] = Field(
        None, title="Working directory to use for the job"
    )


class OpenapiJobInfoResp(BaseModel):
    jobs: Optional[List[OpenapiJobInfoRespJobsItem]] = Field(None, title="List of jobs")
    last_backfill: Optional[OpenapiJobInfoRespLastBackfill] = Field(
        None, title="Time of last backfill scheduler run (UNIX timestamp)"
    )
    last_update: Optional[OpenapiJobInfoRespLastUpdate] = Field(
        None, title="Time of last job change (UNIX timestamp)"
    )
    meta: Optional[OpenapiJobInfoRespMeta] = Field(None, title="Slurm meta values")
    errors: Optional[List[OpenapiJobInfoRespErrorsItem]] = Field(
        None, title="Query errors"
    )
    warnings: Optional[List[OpenapiJobInfoRespWarningsItem]] = Field(
        None, title="Query warnings"
    )


class GetResponse200SlurmJobJobid(BaseModel):
    content: OpenapiJobInfoResp = Field(..., title="Openapi job info resp")


class GetResponse200SlurmJobs(BaseModel):
    content: OpenapiJobInfoResp = Field(..., title="Openapi job info resp")


class GetResponseDefaultSlurmJobJobid(BaseModel):
    content: OpenapiJobInfoResp = Field(..., title="Openapi job info resp")


class GetResponseDefaultSlurmJobs(BaseModel):
    content: OpenapiJobInfoResp = Field(..., title="Openapi job info resp")


class AllModels(BaseModel):
    model0: PostPathSlurmJobSubmit = Field(
        ..., title="POST /slurm/v0.0.41/job/submit: path parameters"
    )
    model1: PostQuerySlurmJobSubmit = Field(
        ..., title="POST /slurm/v0.0.41/job/submit: query parameters"
    )
    model2: JobDescMsgBeginTime = Field(
        ...,
        title="Defer the allocation of the job until the specified time (UNIX timestamp)",
    )
    model3: JobDescMsgCrontabLine = Field(..., title="Job desc msg crontab line")
    model4: JobDescMsgCrontab = Field(..., title="Specification for scrontab job")
    model5: JobDescMsgRlimitsCpu = Field(
        ..., title="Per-process CPU limit, in seconds."
    )
    model6: JobDescMsgRlimitsFsize = Field(
        ..., title="Largest file that can be created, in bytes."
    )
    model7: JobDescMsgRlimitsData = Field(
        ..., title="Maximum size of data segment, in bytes. "
    )
    model8: JobDescMsgRlimitsStack = Field(
        ..., title="Maximum size of stack segment, in bytes."
    )
    model9: JobDescMsgRlimitsCore = Field(
        ..., title="Largest core file that can be created, in bytes."
    )
    model10: JobDescMsgRlimitsRss = Field(
        ...,
        title="Largest resident set size, in bytes. This affects swapping; processes that are exceeding their resident set size will be more likely to have physical memory taken from them.",
    )
    model11: JobDescMsgRlimitsNproc = Field(..., title="Number of processes.")
    model12: JobDescMsgRlimitsNofile = Field(..., title="Number of open files.")
    model13: JobDescMsgRlimitsMemlock = Field(
        ..., title="Locked-in-memory address space"
    )
    model14: JobDescMsgRlimitsAs = Field(..., title="Address space limit.")
    model15: JobDescMsgRlimits = Field(..., title="Job desc msg rlimits")
    model16: JobDescMsgDistributionPlaneSize = Field(
        ..., title="Plane size specification when distribution specifies plane"
    )
    model17: JobDescMsgPriority = Field(..., title="Request specific job priority")
    model18: JobDescMsgTimeLimit = Field(..., title="Maximum run time in minutes")
    model19: JobDescMsgTimeMinimum = Field(..., title="Minimum run time in minutes")
    model20: JobDescMsgKillWarningDelay = Field(
        ..., title="Number of seconds before end time to send the warning signal"
    )
    model21: JobDescMsgMemoryPerCpu = Field(
        ..., title="Minimum memory in megabytes per allocated CPU"
    )
    model22: JobDescMsgMemoryPerNode = Field(
        ..., title="Minimum memory in megabytes per allocated CPU"
    )
    model23: JobDescMsgRequiredSwitches = Field(..., title="Maximum number of switches")
    model24: JobDescMsgSegmentSize = Field(..., title="Segment size for topology/block")
    model25: JobDescMsg = Field(..., title="Job desc msg")
    model26: JobSubmitReq = Field(..., title="Job submit req")
    model27: PostBodySlurmJobSubmit = Field(..., title="Job description")
    model28: OpenapiJobSubmitResponseResult = Field(..., title="Job submission")
    model29: OpenapiJobSubmitResponseMetaPlugin = Field(
        ..., title="Openapi job submit response meta plugin"
    )
    model30: OpenapiJobSubmitResponseMetaClient = Field(
        ..., title="Openapi job submit response meta client"
    )
    model31: OpenapiJobSubmitResponseMetaSlurmVersion = Field(
        ..., title="Openapi job submit response meta slurm version"
    )
    model32: OpenapiJobSubmitResponseMetaSlurm = Field(
        ..., title="Openapi job submit response meta slurm"
    )
    model33: OpenapiJobSubmitResponseMeta = Field(..., title="Slurm meta values")
    model34: OpenapiJobSubmitResponseErrorsItem = Field(
        ..., title="Openapi job submit response errors item"
    )
    model35: OpenapiJobSubmitResponseWarningsItem = Field(
        ..., title="Openapi job submit response warnings item"
    )
    model36: OpenapiJobSubmitResponse = Field(..., title="Openapi job submit response")
    model37: PostResponse200SlurmJobSubmit = Field(..., title="job submission response")
    model38: PostResponseDefaultSlurmJobSubmit = Field(
        ..., title="job submission response"
    )
    model39: GetPathSlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.41/job/{job_id}: path parameters"
    )
    model40: GetQuerySlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.41/job/{job_id}: query parameters"
    )
    model41: GetBodySlurmJobJobid = Field(
        ..., title="GET /slurm/v0.0.41/job/{job_id}: body"
    )
    model42: OpenapiJobInfoRespJobsItemAccrueTime = Field(
        ..., title="When the job started accruing age priority (UNIX timestamp)"
    )
    model43: OpenapiJobInfoRespJobsItemArrayJobId = Field(
        ..., title="Job ID of job array, or 0 if N/A"
    )
    model44: OpenapiJobInfoRespJobsItemArrayTaskId = Field(
        ..., title="Task ID of this task in job array"
    )
    model45: OpenapiJobInfoRespJobsItemArrayMaxTasks = Field(
        ..., title="Maximum number of simultaneously running array tasks, 0 if no limit"
    )
    model46: OpenapiJobInfoRespJobsItemCoresPerSocket = Field(
        ..., title="Cores per socket required"
    )
    model47: OpenapiJobInfoRespJobsItemBillableTres = Field(..., title="Billable TRES")
    model48: OpenapiJobInfoRespJobsItemCpusPerTask = Field(
        ..., title="Number of CPUs required by each task"
    )
    model49: OpenapiJobInfoRespJobsItemCpuFrequencyMinimum = Field(
        ..., title="Minimum CPU frequency"
    )
    model50: OpenapiJobInfoRespJobsItemCpuFrequencyMaximum = Field(
        ..., title="Maximum CPU frequency"
    )
    model51: OpenapiJobInfoRespJobsItemCpuFrequencyGovernor = Field(
        ..., title="CPU frequency governor"
    )
    model52: OpenapiJobInfoRespJobsItemDeadline = Field(
        ..., title="Latest time that the job may start (UNIX timestamp)"
    )
    model53: OpenapiJobInfoRespJobsItemDelayBoot = Field(
        ...,
        title="Number of seconds after job eligible start that nodes will be rebooted to satisfy feature specification",
    )
    model54: OpenapiJobInfoRespJobsItemDerivedExitCodeReturnCode = Field(
        ..., title="Process return code (numeric)"
    )
    model55: OpenapiJobInfoRespJobsItemDerivedExitCodeSignalId = Field(
        ..., title="Signal sent to process (numeric)"
    )
    model56: OpenapiJobInfoRespJobsItemDerivedExitCodeSignal = Field(
        ..., title="Openapi job info resp jobs item derived exit code signal"
    )
    model57: OpenapiJobInfoRespJobsItemDerivedExitCode = Field(
        ..., title="Highest exit code of all job steps"
    )
    model58: OpenapiJobInfoRespJobsItemEligibleTime = Field(
        ..., title="Time when the job became eligible to run (UNIX timestamp)"
    )
    model59: OpenapiJobInfoRespJobsItemEndTime = Field(
        ..., title="End time, real or expected (UNIX timestamp)"
    )
    model60: OpenapiJobInfoRespJobsItemExitCodeReturnCode = Field(
        ..., title="Process return code (numeric)"
    )
    model61: OpenapiJobInfoRespJobsItemExitCodeSignalId = Field(
        ..., title="Signal sent to process (numeric)"
    )
    model62: OpenapiJobInfoRespJobsItemExitCodeSignal = Field(
        ..., title="Openapi job info resp jobs item exit code signal"
    )
    model63: OpenapiJobInfoRespJobsItemExitCode = Field(
        ..., title="Exit code of the job"
    )
    model64: OpenapiJobInfoRespJobsItemHetJobId = Field(
        ..., title="Heterogeneous job ID, if applicable"
    )
    model65: OpenapiJobInfoRespJobsItemHetJobOffset = Field(
        ...,
        title="Unique sequence number applied to this component of the heterogeneous job",
    )
    model66: OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemCpus = Field(
        ...,
        title="Openapi job info resp jobs item job resources nodes allocation item cpus",
    )
    model67: OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemMemory = Field(
        ...,
        title="Openapi job info resp jobs item job resources nodes allocation item memory",
    )
    model68: (
        OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItemCoresItem
    ) = Field(
        ...,
        title="Openapi job info resp jobs item job resources nodes allocation item sockets item cores item",
    )
    model69: OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItemSocketsItem = (
        Field(
            ...,
            title="Openapi job info resp jobs item job resources nodes allocation item sockets item",
        )
    )
    model70: OpenapiJobInfoRespJobsItemJobResourcesNodesAllocationItem = Field(
        ..., title="Job resources for a node"
    )
    model71: OpenapiJobInfoRespJobsItemJobResourcesNodes = Field(
        ..., title="Openapi job info resp jobs item job resources nodes"
    )
    model72: OpenapiJobInfoRespJobsItemJobResourcesThreadsPerCore = Field(
        ..., title="Number of processor threads per CPU core"
    )
    model73: OpenapiJobInfoRespJobsItemJobResources = Field(
        ..., title="Resources used by the job"
    )
    model74: OpenapiJobInfoRespJobsItemLastSchedEvaluation = Field(
        ..., title="Last time job was evaluated for scheduling (UNIX timestamp)"
    )
    model75: OpenapiJobInfoRespJobsItemMaxCpus = Field(
        ..., title="Maximum number of CPUs usable by the job"
    )
    model76: OpenapiJobInfoRespJobsItemMaxNodes = Field(
        ..., title="Maximum number of nodes usable by the job"
    )
    model77: OpenapiJobInfoRespJobsItemTasksPerCore = Field(
        ..., title="Number of tasks invoked on each core"
    )
    model78: OpenapiJobInfoRespJobsItemTasksPerTres = Field(
        ..., title="Number of tasks that can assess each GPU"
    )
    model79: OpenapiJobInfoRespJobsItemTasksPerNode = Field(
        ..., title="Number of tasks invoked on each node"
    )
    model80: OpenapiJobInfoRespJobsItemTasksPerSocket = Field(
        ..., title="Number of tasks invoked on each socket"
    )
    model81: OpenapiJobInfoRespJobsItemTasksPerBoard = Field(
        ..., title="Number of tasks invoked on each board"
    )
    model82: OpenapiJobInfoRespJobsItemCpus = Field(
        ..., title="Minimum number of CPUs required"
    )
    model83: OpenapiJobInfoRespJobsItemNodeCount = Field(
        ..., title="Minimum number of nodes required"
    )
    model84: OpenapiJobInfoRespJobsItemTasks = Field(..., title="Number of tasks")
    model85: OpenapiJobInfoRespJobsItemMemoryPerCpu = Field(
        ..., title="Minimum memory in megabytes per allocated CPU"
    )
    model86: OpenapiJobInfoRespJobsItemMemoryPerNode = Field(
        ..., title="Minimum memory in megabytes per allocated node"
    )
    model87: OpenapiJobInfoRespJobsItemMinimumCpusPerNode = Field(
        ..., title="Minimum number of CPUs per node"
    )
    model88: OpenapiJobInfoRespJobsItemMinimumTmpDiskPerNode = Field(
        ..., title="Minimum tmp disk space required per node"
    )
    model89: OpenapiJobInfoRespJobsItemPower = Field(
        ..., title="Openapi job info resp jobs item power"
    )
    model90: OpenapiJobInfoRespJobsItemPreemptTime = Field(
        ..., title="Time job received preemption signal (UNIX timestamp)"
    )
    model91: OpenapiJobInfoRespJobsItemPreemptableTime = Field(
        ..., title="Time job becomes eligible for preemption (UNIX timestamp)"
    )
    model92: OpenapiJobInfoRespJobsItemPreSusTime = Field(
        ..., title="Total run time prior to last suspend in seconds"
    )
    model93: OpenapiJobInfoRespJobsItemPriority = Field(
        ..., title="Request specific job priority"
    )
    model94: OpenapiJobInfoRespJobsItemResizeTime = Field(
        ..., title="Time of last size change (UNIX timestamp)"
    )
    model95: OpenapiJobInfoRespJobsItemSocketsPerNode = Field(
        ..., title="Number of sockets per node required"
    )
    model96: OpenapiJobInfoRespJobsItemStartTime = Field(
        ..., title="Time execution began, or is expected to begin (UNIX timestamp)"
    )
    model97: OpenapiJobInfoRespJobsItemSubmitTime = Field(
        ..., title="Time when the job was submitted (UNIX timestamp)"
    )
    model98: OpenapiJobInfoRespJobsItemSuspendTime = Field(
        ..., title="Time the job was last suspended or resumed (UNIX timestamp)"
    )
    model99: OpenapiJobInfoRespJobsItemTimeLimit = Field(
        ..., title="Maximum run time in minutes"
    )
    model100: OpenapiJobInfoRespJobsItemTimeMinimum = Field(
        ..., title="Minimum run time in minutes"
    )
    model101: OpenapiJobInfoRespJobsItemThreadsPerCore = Field(
        ..., title="Number of processor threads per CPU core required"
    )
    model102: OpenapiJobInfoRespJobsItem = Field(
        ..., title="Openapi job info resp jobs item"
    )
    model103: OpenapiJobInfoRespLastBackfill = Field(
        ..., title="Time of last backfill scheduler run (UNIX timestamp)"
    )
    model104: OpenapiJobInfoRespLastUpdate = Field(
        ..., title="Time of last job change (UNIX timestamp)"
    )
    model105: OpenapiJobInfoRespMetaPlugin = Field(
        ..., title="Openapi job info resp meta plugin"
    )
    model106: OpenapiJobInfoRespMetaClient = Field(
        ..., title="Openapi job info resp meta client"
    )
    model107: OpenapiJobInfoRespMetaSlurmVersion = Field(
        ..., title="Openapi job info resp meta slurm version"
    )
    model108: OpenapiJobInfoRespMetaSlurm = Field(
        ..., title="Openapi job info resp meta slurm"
    )
    model109: OpenapiJobInfoRespMeta = Field(..., title="Slurm meta values")
    model110: OpenapiJobInfoRespErrorsItem = Field(
        ..., title="Openapi job info resp errors item"
    )
    model111: OpenapiJobInfoRespWarningsItem = Field(
        ..., title="Openapi job info resp warnings item"
    )
    model112: OpenapiJobInfoResp = Field(..., title="Openapi job info resp")
    model113: GetResponse200SlurmJobJobid = Field(..., title="job(s) information")
    model114: GetResponseDefaultSlurmJobJobid = Field(..., title="job(s) information")
    model115: DeletePathSlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.41/job/{job_id}: path parameters"
    )
    model116: DeleteQuerySlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.41/job/{job_id}: query parameters"
    )
    model117: DeleteBodySlurmJobJobid = Field(
        ..., title="DELETE /slurm/v0.0.41/job/{job_id}: body"
    )
    model118: OpenapiRespMetaPlugin = Field(..., title="Openapi resp meta plugin")
    model119: OpenapiRespMetaClient = Field(..., title="Openapi resp meta client")
    model120: OpenapiRespMetaSlurmVersion = Field(
        ..., title="Openapi resp meta slurm version"
    )
    model121: OpenapiRespMetaSlurm = Field(..., title="Openapi resp meta slurm")
    model122: OpenapiRespMeta = Field(..., title="Slurm meta values")
    model123: OpenapiRespErrorsItem = Field(..., title="Openapi resp errors item")
    model124: OpenapiRespWarningsItem = Field(..., title="Openapi resp warnings item")
    model125: OpenapiResp = Field(..., title="Openapi resp")
    model126: DeleteResponse200SlurmJobJobid = Field(..., title="job signal result")
    model127: DeleteResponseDefaultSlurmJobJobid = Field(..., title="job signal result")
    model128: GetPathSlurmJobs = Field(
        ..., title="GET /slurm/v0.0.41/jobs: path parameters"
    )
    model129: GetQuerySlurmJobs = Field(
        ..., title="GET /slurm/v0.0.41/jobs: query parameters"
    )
    model130: GetBodySlurmJobs = Field(..., title="GET /slurm/v0.0.41/jobs: body")
    model131: GetResponse200SlurmJobs = Field(..., title="job(s) information")
    model132: GetResponseDefaultSlurmJobs = Field(..., title="job(s) information")



import os
import threading
from typing import Generator, List
from contextlib import contextmanager

import pytest

from ..client.job_io.local import Connection
from ..client.job_io.local import TcpConnection
from ..client.job_io.local import FileConnection
from ..client.job_io.local import LocalWorkerProxy
from ..client.job_io.local import RemoteWorkerProxy

from ..client.job_io.remote import tcp_main
from ..client.job_io.remote import file_main


def test_single_task(communication):
    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy()
    else:
        proxy_ctx = RemoteWorkerProxy(connection)

    with proxy_ctx as proxy:
        assert proxy.execute(sum, ([1, 2],), None) == 3


def test_single_task_exception(communication):
    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy()
    else:
        proxy_ctx = RemoteWorkerProxy(connection)

    with proxy_ctx as proxy:
        with pytest.raises(TypeError):
            proxy.execute(sum, ([1, "2"],), None)


def test_max_tasks(communication):
    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy(max_tasks=2)
    else:
        proxy_ctx = RemoteWorkerProxy(connection, max_tasks=2)

    with proxy_ctx as proxy:
        assert proxy.execute(sum, ([1, 2],), None) == 3
        assert proxy.execute(sum, ([3, 4],), None) == 7


def test_max_tasks_exception(communication):
    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy(max_tasks=2)
    else:
        proxy_ctx = RemoteWorkerProxy(connection, max_tasks=2)

    with proxy_ctx as proxy:
        assert proxy.execute(sum, ([1, 2],), None) == 3
        with pytest.raises(TypeError):
            proxy.execute(sum, ([1, "2"],), None)
        with pytest.raises(RuntimeError, match="cannot send data after stopped"):
            proxy.execute(sum, ([3, 4],), None)


def test_no_max_tasks(communication):
    connection, remote = communication
    if connection is None:
        pytest.skip("test needs a remote")

    proxy_ctx = RemoteWorkerProxy(connection, max_tasks=None)

    with proxy_ctx as proxy:
        assert proxy.execute(sum, ([1, 2],), None) == 3
        remote.join(timeout=1)
        assert remote.is_alive()


def test_initializer(communication):
    global _INITIALIZED
    _INITIALIZED = None

    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy(initializer=_initializer, initargs=([10],))
    else:
        proxy_ctx = RemoteWorkerProxy(
            connection, initializer=_initializer, initargs=([10],)
        )

    with proxy_ctx as proxy:
        assert proxy.execute(_sum_to_initialized, ([1, 2],), None) == 13


def _initializer(value: List[int]) -> None:
    global _INITIALIZED
    try:
        if isinstance(_INITIALIZED, Exception):
            raise _INITIALIZED
    except NameError:
        pass
    _INITIALIZED = value


def _sum_to_initialized(value: List[int]) -> int:
    global _INITIALIZED  # noqa F824
    return sum(value + _INITIALIZED)


def test_initializer_exception(communication):
    global _INITIALIZED
    _INITIALIZED = RuntimeError("intentional for testing")

    connection, _ = communication
    if connection is None:
        proxy_ctx = LocalWorkerProxy(initializer=_failing_initializer)
    else:
        proxy_ctx = RemoteWorkerProxy(connection, initializer=_failing_initializer)

    with proxy_ctx as proxy:
        with pytest.raises(RuntimeError, match="intentional for testing"):
            proxy.initialize()
        with pytest.raises(RuntimeError, match="cannot send data after stopped"):
            proxy.execute(sum, ([1, 2],), None)


def _failing_initializer() -> None:
    raise RuntimeError("intentional for testing")


@pytest.fixture(params=["tcp", "file", "local"])
def communication(request, tmp_path):
    if request.param == "tcp":
        with _tcp_connection() as conn:
            with _remote_tcp_env(conn):
                with _remote_job(tcp_main) as remote:
                    yield conn, remote
    elif request.param == "file":
        with _file_connection(tmp_path) as conn:
            with remote_file_env(conn):
                with _remote_job(file_main) as remote:
                    yield conn, remote
    else:
        yield None, None


@contextmanager
def _tcp_connection() -> Generator[Connection, None, None]:
    """Start the client-side TCP-based connection."""
    with TcpConnection() as conn:
        yield conn


@contextmanager
def _file_connection(tmp_path) -> Generator[Connection, None, None]:
    """Start the client-side file-based connection."""
    with FileConnection(str(tmp_path), "test") as conn:
        yield conn


@contextmanager
def _remote_tcp_env(connection: TcpConnection) -> Generator[None, None, None]:
    """Setup environment for the remote job."""
    os.environ["_PYSLURMUTILS_HOST"] = connection.host
    os.environ["_PYSLURMUTILS_PORT"] = str(connection.port)
    try:
        yield
    finally:
        del os.environ["_PYSLURMUTILS_HOST"]
        del os.environ["_PYSLURMUTILS_PORT"]


@contextmanager
def remote_file_env(connection: TcpConnection) -> Generator[None, None, None]:
    """Setup environment for the remote job."""
    os.environ["_PYSLURMUTILS_INFILE"] = connection.input_filename
    os.environ["_PYSLURMUTILS_OUTFILE"] = connection.output_filename
    try:
        yield
    finally:
        del os.environ["_PYSLURMUTILS_INFILE"]
        del os.environ["_PYSLURMUTILS_OUTFILE"]


@contextmanager
def _remote_job(remote_main: callable) -> Generator[threading.Thread, None, None]:
    """Run the remote main function in a local thread."""
    thread = threading.Thread(target=remote_main, daemon=True)
    thread.start()
    try:
        yield thread
    finally:
        thread.join(timeout=60)
        assert not thread.is_alive()



import logging
import importlib
from typing import Optional, Dict, List, Any, Type, Tuple

import requests
from requests.exceptions import HTTPError
from requests.exceptions import InvalidJSONError
from pydantic import BaseModel
from pydantic import ValidationError

from . import slurm_response
from . import slurm_endpoints
from ...errors import RemoteHttpError
from ...errors import RemoteSlurmError
from ...errors import raise_chained_errors

logger = logging.getLogger(__name__)


def validated_slurm_request(
    method: str,
    base_url: str,
    path: str,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
    json: Optional[Dict] = None,
    headers: Optional[Dict] = None,
    raise_http_error: bool = True,
    **kwargs,
) -> BaseModel:
    """
    Sends an HTTP request and convert the response to the appropriate Pydantic model.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param base_url: The base URL to which the path is appended.
    :param path: The specific API path for the request (e.g., "/slurm/v0.0.38/job/{job_id}").
    :param path_params: A dictionary of path parameters to be substituted in the path.
    :param query_params: A dictionary of query parameters.
    :param json: A dictionary to represent the JSON body for POST/PUT requests.
    :param headers: A dictionary to represent the header for the request.
    :param raise_http_error: Raise `RemoteHttpError`.
    :param kwargs: Additional parameters to be passed to the `requests.request()` call.
    :return: A Pydantic model instance of the HTTP request response.
    :raises ValueError: If header keys are missing or response validation failed.
    :raises RemoteHttpError: The HTTP request status indicates an error.
    """
    if path_params:
        path_params = {k: str(v) for k, v in path_params.items()}
        url = base_url + path.format(**path_params)
    else:
        url = base_url + path

    if query_params:
        query_params = {k: str(v) for k, v in query_params.items()}

    _request_validate_headers(method, path, headers=headers)

    _warn_invalid_request_parameters(
        method, path, path_params=path_params, query_params=query_params, body=json
    )

    response = requests.request(
        method, url, params=query_params, json=json, headers=headers, **kwargs
    )

    response_model = _request_parse_response(method, path, response)

    _request_response_handling(
        method, url, response, response_model, raise_http_error=raise_http_error
    )

    return response_model


def slurm_request(
    method: str, base_url: str, path: str, headers: Optional[Dict] = None, **kwargs
) -> Any:
    """
    Sends an HTTP request and JSON decode the request.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param base_url: The base URL to which the path is appended.
    :param path: The specific API path for the request (e.g., "/openapi").
    :param headers: A dictionary to represent the header for the request.
    :param kwargs: Additional parameters to be passed to the `requests.request()` call.
    :return: JSON decoded result.
    :raises ValueError: If header keys are missing or response decoding failed.
    :raises RemoteHttpError: The HTTP request status indicates an error.
    """
    _request_validate_headers(method, path, headers=headers)

    url = base_url + path
    response = requests.request(method, url, headers=headers, **kwargs)

    try:
        response.raise_for_status()
    except HTTPError as e:
        raise RemoteHttpError(f"{method} {path} failed") from e

    try:
        return response.json()
    except InvalidJSONError:
        raise ValueError(
            f"{method} {path} response decoding failed: {response.text}"
        ) from None


def parse_version(version: str) -> Tuple[int, int, int]:
    """
    Parse a version string into a tuple of integers.

    :param version: A version string in the format 'vX.X.X'.
    :return: A tuple representing the version (X, X, X).
    :raises ValueError: When the version string is badly formatted.
    """
    try:
        version_tuple = tuple(map(int, version.lstrip("v").split(".")))
        assert len(version_tuple) == 3
        return version_tuple
    except (ValueError, TypeError, AssertionError):
        raise ValueError(
            f"Invalid slurm version '{version}'. The format must be 'vX.X.X'"
        ) from None


def create_version(version: Tuple[int, int, int]) -> str:
    """
    Create a version string from a tuple of integers.

    :param version: A tuple representing the version (X, X, X).
    :return: A version string in the format 'vX.X.X'.
    """
    return f"v{'.'.join(map(str, version))}"


def _request_validate_headers(
    method: str,
    path: str,
    headers: Optional[Dict] = None,
) -> None:
    """
    Sends an HTTP request and converts the response to the appropriate Pydantic model.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param path: The specific API path for the request (e.g., "/slurm/v0.0.38/job/{job_id}").
    :param headers: The request headers.
    :raises ValueError: If header keys are missing.
    """
    if not headers:
        raise ValueError(f"{method} {path}: No header provided")
    for key in ("X-SLURM-USER-NAME", "X-SLURM-USER-TOKEN"):
        if not headers.get(key):
            raise ValueError(f"{method} {path}: Missing header key '{key}'")


def _warn_invalid_request_parameters(
    method: str,
    path: str,
    path_params: Optional[Dict[str, str]],
    query_params: Optional[Dict[str, str]],
    body: Optional[Dict],
) -> None:
    """
    Validate HTTP request parameters and log warning when not valid.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param path: The specific API path for the request (e.g., "/slurm/v0.0.38/job/{job_id}").
    :param path_params: A dictionary of path parameters to be substituted in the path.
    :param query_params: A dictionary of query parameters.
    :param body: A dictionary to represent the JSON body for POST/PUT requests.
    """

    model_class = _load_request_model(method, path, "path")
    _warn_invalid_data(
        path_params,
        model_class,
        f"{method} {path}: invalid path parameters. Submit request anyway.",
    )

    model_class = _load_request_model(method, path, "query")
    _warn_invalid_data(
        query_params,
        model_class,
        f"{method} {path}: invalid query parameters. Submit request anyway.",
    )

    model_class = _load_request_model(method, path, "body")
    _warn_invalid_data(
        {"content": body},
        model_class,
        f"{method} {path}: invalid request body. Submit request anyway.",
    )


def _warn_invalid_data(data: Any, model_class: Type[BaseModel], message: str) -> None:
    """
    Validate data and log warning when not valid.

    :param data: Data to be validated
    :param model_class: Pydantic model with which to validate the data.
    :param message: Warning message in addition to the validation errors.
    """
    try:
        if data is None:
            _ = model_class()
        else:
            _ = model_class.model_validate(data)
    except ValidationError as e:
        logger.warning(_compose_validation_message(e, message))


def _request_parse_response(
    method: str,
    path: str,
    response: requests.Response,
) -> BaseModel:
    """
    Sends an HTTP request and converts the response to the appropriate Pydantic model.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param path: The specific API path for the request (e.g., "/slurm/v0.0.38/job/{job_id}").
    :param response: Response from HTTP request.
    :return: A Pydantic model instance of the HTTP request response.
    :raises ValueError: Pydantic model validation failed.
    """
    response_class = _load_request_model(
        method, path, str(response.status_code), "default"
    )

    # JSON decoding

    try:
        response_data = response.json()
    except InvalidJSONError:
        raise ValueError(
            f"{method} {path} response decoding failed: {response.text}"
        ) from None

    data_to_validate = {"content": response_data}

    # Pydantic validation (validation failure allowed)

    try:
        return response_class.model_validate(data_to_validate).content
    except ValidationError as e:
        message = f"{method} {path}: invalid response. Try removing invalid items."
        logger.warning(_compose_validation_message(e, message))
        validation_errors = e.errors()

    # Remove invalid items.
    #
    # Response models have only optional fields (see `force_optional=True`
    # in the model generation script) despite some being required
    # in the OpenAPI spec. We do this for best effort response handling.
    # Any field could be abscent and an AttributeError will be raised.

    data_to_validate = _remove_invalid_items(data_to_validate, validation_errors)

    # Pydantic validation (no validation failure allowed)
    try:
        return response_class.model_validate(data_to_validate).content
    except ValidationError as e:
        message = f"{method} {path}: invalid response."
        raise ValueError(_compose_validation_message(e, message)) from None


def _compose_validation_message(exc: ValidationError, message: str) -> str:
    validation_error = str(exc).replace("\n", "\n ")
    return f"{message}\n {validation_error}\n"


def _remove_invalid_items(data: Any, validation_errors: List[Dict[str, Any]]) -> Any:
    """
    Removes invalid items from the data based on the error details.

    :param data: The data that may contain invalid items (usually the parsed response).
    :param validation_errors: The list of error details returned by Pydantic.
    :return: The data with invalid items removed.
    """
    invalid_paths = {tuple(error["loc"]): error for error in validation_errors}
    return _recursive_remove(data, (), invalid_paths)


def _recursive_remove(data: Any, path: tuple, invalid_paths: set) -> Any:
    """
    Recursively removes invalid items from nested data.

    :param data: The data (dict/list) to process.
    :param path: The path to the field to remove.
    :param invalid_paths: A set of paths that should be removed.
    :return: The data with invalid items removed.
    """
    if isinstance(data, dict):
        data_copy = {}
        for key, value in data.items():
            new_path = path + (key,)
            if new_path in invalid_paths:
                logger.debug(
                    "Remove key %s from path %s", key, ".".join(map(str, path))
                )
            else:
                data_copy[key] = _recursive_remove(value, new_path, invalid_paths)
        return data_copy

    if isinstance(data, list):
        data_copy = []
        for index in range(len(data)):
            new_path = path + (index,)
            if new_path in invalid_paths:
                logger.debug(
                    "Remove index %s from path %s", index, ".".join(map(str, path))
                )
            else:
                data_copy.append(
                    _recursive_remove(data[index], new_path, invalid_paths)
                )
        return data_copy

    return data


def _request_response_handling(
    method: str,
    path: str,
    response: requests.Response,
    response_model: BaseModel,
    raise_http_error: str,
):
    """
    Handle the HTTP response.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param url: HTTP URL.
    :param response: Response from HTTP request.
    :param response_model: Pydantic model of the response data.
    :param raise_http_error: Raise `RemoteHttpError`.
    :raises RemoteHttpError: The HTTP request status indicates an error.
    """
    for warning in slurm_response.slurm_warning_messages(method, path, response_model):
        logger.warning(warning)

    errors, suffix = slurm_response.slurm_error_messages(response_model)
    errors = [RemoteSlurmError(msg) for msg in errors]

    try:
        response.raise_for_status()
    except HTTPError as e:
        errors.append(e)

    if errors:
        if raise_http_error:
            errors.append(RemoteHttpError(f"{method} {path} failed{suffix}"))
            raise_chained_errors(errors)
        else:
            logger.error(
                "%s %s failed%s\n %s",
                method,
                path,
                suffix,
                "\n ".join(map(str, errors)),
            )


def _load_request_model(
    method: str, path: str, field: str, default_field: Optional[str] = None
) -> Type[BaseModel]:
    """
    Load the Pydantic model class for an endpoint field.

    :param method: HTTP method to use for the request (e.g., "GET", "POST").
    :param path: The specific API path for the request (e.g., "/slurm/v0.0.38/job/{job_id}").
    :param field: Endpoint field name.
    :param default_field: Default field name when it does not exist.
    :return: A Pydantic model class.
    :raises KeyError: Endpoint or endpoint field is not supported.
    """
    try:
        module_name, model_to_class_name = slurm_endpoints.ENDPOINTS[(method, path)]
    except KeyError:
        raise KeyError(f"{method} {path} is not supported") from None

    if field not in model_to_class_name and default_field:
        field = default_field

    try:
        class_name = model_to_class_name[field]
    except KeyError:
        raise KeyError(
            f"Field '{field}' is not supported for {method} {path}"
        ) from None

    module = _load_module(module_name)
    return getattr(module, class_name)


def _load_module(module_name: str) -> BaseModel:
    """
    :param module_name:
    :raises KeyError: Endpoint is not supported.
    """
    return importlib.import_module(f".{module_name}", package=__package__)
